"""
Common fixtures used by many tests.
"""

import os
import yaml
import pytest


from lute.parse.registry import init_parser_plugins

from lute.models.language import Language


def pytest_sessionstart(session):  # pylint: disable=unused-argument
    """
    Initialize parser list
    """
    init_parser_plugins()


def _get_test_language():
    """
    Retrieve the language definition file for testing ths plugin from definition.yaml
    """
    thisdir = os.path.dirname(os.path.realpath(__file__))
    definition_file = os.path.join(thisdir, "..", "definition.yaml")
    with open(definition_file, "r", encoding="utf-8") as df:
        d = yaml.safe_load(df)
    lang = Language.from_dict(d)
    return lang


@pytest.fixture(name="khmer")
def fixture_khmer():
    return _get_test_language()



"Smoke test only."

from lute.cli.language_term_export import generate_language_file, generate_book_file

from lute.models.term import Term, TermTag
from lute.models.repositories import TermRepository, LanguageRepository
from lute.models.book import Book
from lute.db import db
from lute.db.demo import Service as DemoService
from tests.dbasserts import assert_sql_result, assert_record_count_equals


def test_language_term_export_smoke_test(app_context, tmp_path):
    "dump data."
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    sql = """select * from books
      where BkLgID = (select LgID from languages where LgName='English')
    """
    assert_record_count_equals(sql, 2, "have books")
    langrepo = LanguageRepository(db.session)
    eng = langrepo.find_by_name("English")
    t = Term(eng, "the")
    t.translation = "article"
    t.add_term_tag(TermTag("a"))
    t.add_term_tag(TermTag("b"))
    db.session.add(t)
    db.session.commit()

    outfile = tmp_path / "outfile.csv"
    generate_language_file("English", outfile)
    with open(outfile, "r", encoding="utf-8") as ofhandle:
        text = ofhandle.read()
    print(text)
    lines = text.split("\n")
    head = lines[0]
    assert (
        head == "term,count,familycount,books,definition,status,parents,children,tags"
    ), "headings"
    firstline = lines[1]
    assert firstline.startswith("the,"), "the is most common"
    assert firstline.endswith('article,1,-,-,"a, b"'), "ending data"


def test_single_book_export(app_context, empty_db, tmp_path, english):
    "dump data for english."

    assert_sql_result("select * from books", [], "no books")
    assert_sql_result("select * from words", [], "no terms")

    fulltext = "a b c d e A B C\n---\nG H I c d e d"
    b = Book.create_book("hi", english, fulltext)
    db.session.add(b)
    db.session.commit()

    for c in ["a", "d", "c d"]:
        t = Term(english, c)
        t.status = 1
        db.session.add(t)
    for c in ["e", "g", "h"]:
        t = Term(english, c)
        t.status = 0
        db.session.add(t)
    db.session.commit()

    def _find(term_string):
        "Find term with the text."
        spec = Term(english, term_string)
        repo = TermRepository(db.session)
        ret = repo.find_by_spec(spec)
        assert ret is not None, f"Have {term_string}"
        return ret

    a = _find("a")
    for c in ["e", "h"]:
        t = _find(c)
        t.add_parent(a)
        db.session.add(t)
    db.session.commit()

    expected = [
        # Headings
        "term,count,familycount,books,definition,status,parents,children,tags",
        # a has two children, e and h
        "a,2,5,hi,-,1,-,e (2); h (1),-",
        # b occurs twice.
        "b,2,2,hi,-,0,-,-,-",
        # 'c d' occurs twice
        "c d,2,2,hi,-,1,-,-,-",
        # e is a new term
        "e,2,2,hi,-,0,a,-,-",
        # c is a new term, status 0.
        # Occurs once as c, once as C.
        "C,1,1,hi,-,0,-,-,-",
        "I,1,1,hi,-,0,-,-,-",
        "d,1,1,hi,-,1,-,-,-",
        # g and h are new
        "g,1,1,hi,-,0,-,-,-",
        "h,1,1,hi,-,0,a,-,-",
        "",
    ]

    def _lowersort(arr):
        "Lowercase and sort strings."
        return sorted([a.lower() for a in arr])

    def _assert_text_matches_expected(file_text, expected_array):
        "Avoid sorting, case issues."

        # Converting the text to lower because sometimes the text file
        # returned B, and sometimes b ...  which is _very_ odd, but I
        # don't really care.
        assert _lowersort(file_text.split("\n")) == _lowersort(expected_array)

    # Generate for english.
    outfile = tmp_path / "outfile.csv"
    generate_language_file("English", outfile)
    with open(outfile, "r", encoding="utf-8") as ofhandle:
        text = ofhandle.read()
    _assert_text_matches_expected(text, expected)

    # Generate for book.
    generate_book_file(b.id, outfile)
    with open(outfile, "r", encoding="utf-8") as ofhandle:
        text = ofhandle.read()
    _assert_text_matches_expected(text, expected)



"""
Term status following.
"""

import pytest

from lute.models.term import Term as DBTerm
from lute.db import db
from tests.dbasserts import assert_sql_result


@pytest.fixture(name="term_family")
def fixture_term_family(app_context, english):
    """
    Term family.

    A
      B - follows
        b1 - follows
        b2 - no
      C - does not follow
        c1 - follows
        c2 - no
    """

    class family:
        "Family of terms."

        def __init__(self):
            "Set up terms."
            # pylint: disable=invalid-name
            A = DBTerm(english, "A")
            B = DBTerm(english, "Byes")
            B.add_parent(A)
            B.sync_status = True
            b1 = DBTerm(english, "b1yes")
            b1.add_parent(B)
            b1.sync_status = True
            b2 = DBTerm(english, "b2no")
            b2.add_parent(B)

            C = DBTerm(english, "Cno")
            C.add_parent(A)
            c1 = DBTerm(english, "c1yes")
            c1.add_parent(C)
            c1.sync_status = True
            c2 = DBTerm(english, "c2no")
            c2.add_parent(C)

            db.session.add(A)
            db.session.add(B)
            db.session.add(b1)
            db.session.add(b2)
            db.session.add(C)
            db.session.add(c1)
            db.session.add(c2)
            db.session.commit()

            self.A = A
            self.B = B
            self.b1 = b1
            self.b2 = b2
            self.C = C
            self.c1 = c1
            self.c2 = c2

    f = family()

    expected_initial_state = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 1
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected_initial_state, "initial state")

    return f


def assert_statuses(expected, msg):
    "Check the statuses of terms."
    lines = [
        s.strip().replace(":", ";") for s in expected.split("\n") if s.strip() != ""
    ]
    sql = "select WoText, WoStatus from words order by WoTextLC"
    assert_sql_result(sql, lines, msg)


def test_parent_status_propagates_down(term_family, app_context):
    "Changing status should propagate down the tree."
    f = term_family
    f.A.status = 4
    db.session.add(f.A)
    db.session.commit()

    expected = """
    A: 4
    Byes: 4
    b/1/yes: 4
    b/2/no: 1
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_status_propagates_up_and_down(term_family, app_context):
    "Parent and child also updated."
    f = term_family
    f.B.status = 4
    db.session.add(f.B)
    db.session.commit()

    expected = """
    A: 4
    Byes: 4
    b/1/yes: 4
    b/2/no: 1
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_status_propagates_even_if_circular_reference(term_family, app_context):
    "B-b1-B circular reference, is all ok."
    f = term_family
    # f.b1.parents.append(f.B)
    f.B.parents.append(f.b1)
    db.session.add(f.B)
    # db.session.add(f.b1)
    db.session.commit()

    zws = "\u200B"
    sql = f"""
    select p.WoText, c.WoText
    from words p
    inner join wordparents on p.WoID = WpParentWoID
    inner join words c on c.WoID = WpWoID
    where p.WoText in ('Byes', 'b{zws}1{zws}yes')
    """
    assert_sql_result(
        sql, ["Byes; b/1/yes", "Byes; b/2/no", "b/1/yes; Byes"], "parent set"
    )

    f.B.status = 4
    db.session.add(f.B)
    db.session.commit()
    expected = """
    A: 1
    Byes: 4
    b/1/yes: 4
    b/2/no: 1
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_status_stops_propagating_to_top(term_family, app_context):
    "Goes up the tree until it stops."
    f = term_family
    f.c1.status = 4
    db.session.add(f.c1)
    db.session.commit()

    expected = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 1
    Cno: 4
    c/1/yes: 4
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_term_not_following_parent(term_family, app_context):
    "Doesn't update parent."
    f = term_family
    f.b2.status = 4
    db.session.add(f.b2)
    db.session.commit()

    expected = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 4
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_changing_sync_status_propagates_status(term_family, app_context):
    "Doesn't update parent."
    f = term_family
    f.b2.status = 4
    db.session.add(f.b2)
    db.session.commit()

    expected = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 4
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "not following, not updated")

    f.b2.sync_status = True
    db.session.add(f.b2)
    db.session.commit()

    expected = """
    A: 4
    Byes: 4
    b/1/yes: 4
    b/2/no: 4
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_parent_not_updated_if_term_has_multiple_parents(term_family, app_context):
    "Doesn't update parent."
    f = term_family
    f.c1.add_parent(f.B)
    db.session.add(f.c1)
    db.session.commit()

    f.c1.status = 4
    db.session.add(f.c1)
    db.session.commit()

    expected = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 1
    Cno: 1
    c/1/yes: 4
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_adding_new_term_changes_family_if_added(english, term_family, app_context):
    "Doesn't update parent."
    f = term_family

    b3 = DBTerm(english, "b3yes")
    b3.add_parent(f.B)
    b3.status = 3
    b3.sync_status = True
    db.session.add(b3)
    db.session.commit()

    expected = """
    A: 3
    Byes: 3
    b/1/yes: 3
    b/2/no: 1
    b/3/yes: 3
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def test_adding_new_term_does_not_change_family_if_multiple_parents(
    english, term_family, app_context
):
    "Doesn't update parent."
    f = term_family

    b3 = DBTerm(english, "b3yes")
    b3.parents.append(f.B)
    b3.parents.append(f.C)
    b3.status = 3
    # NOTE: setting "follow parent" causes weird propagation ...
    db.session.add(b3)
    db.session.commit()

    expected = """
    A: 1
    Byes: 1
    b/1/yes: 1
    b/2/no: 1
    b/3/yes: 3
    Cno: 1
    c/1/yes: 1
    c/2/no: 1
    """
    assert_statuses(expected, "updated")


def assert_sync_statuses(expected, msg):
    "Check the sync statuses of terms."
    lines = [
        s.strip().replace(":", ";") for s in expected.split("\n") if s.strip() != ""
    ]
    sql = "select WoText from words where WoSyncStatus = 1 order by WoText"
    assert_sql_result(sql, lines, msg)


def test_deleting_parent_deactivates_sync_status(term_family, app_context):
    "No more parent = no more follow."

    expected = """
    Byes
    b/1/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "before delete")

    f = term_family
    db.session.delete(f.A)
    db.session.commit()

    expected = """
    b/1/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "after delete")


def test_adding_extra_parents_unsets_sync_status(term_family, app_context):
    "Can't follow multiple people."
    c1 = db.session.get(DBTerm, term_family.c1.id)
    assert c1.sync_status is True, "following C"

    c1.add_parent(term_family.B)
    assert c1.sync_status is False, "Can't follow 2 parents"


def test_changing_parent_keeps_sync_status(term_family, app_context):
    "Can't follow multiple people."
    c1 = db.session.get(DBTerm, term_family.c1.id)
    assert c1.sync_status is True, "following C"

    c1.remove_all_parents()
    c1.add_parent(term_family.B)
    assert c1.sync_status is True, "Still following"

    db.session.add(c1)
    db.session.commit()

    expected = """
    Byes
    b/1/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "after change")


def test_remove_parent_doesnt_affect_other_linked_terms(term_family, app_context):
    "Issue 417."

    b3 = DBTerm(term_family.B.language, "b3yes")
    b3.add_parent(term_family.B)
    b3.sync_status = True
    db.session.add(b3)
    db.session.commit()

    expected = """
    Byes
    b/1/yes
    b/3/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "set up")

    b1 = term_family.b1
    b1.remove_all_parents()
    b1.sync_status = False
    db.session.add(b1)
    db.session.commit()

    expected = """
    Byes
    b/3/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "set up")


def test_remove_non_linked_parent_leaves_other_linked_terms(term_family, app_context):
    "Issue 417."

    # term b2 is not linked to its parent b
    b2 = term_family.b2
    b2.remove_all_parents()
    b2.sync_status = False
    db.session.add(b2)
    db.session.commit()

    expected = """
    Byes
    b/1/yes
    c/1/yes
    """
    assert_sync_statuses(expected, "b1 still linked")



"""
Language model tests - getting, saving, etc.

Low value but ensure that the db mapping is correct.
"""

from lute.db import db
from lute.db.demo import Service as DemoService
from lute.models.language import Language
from lute.models.repositories import LanguageRepository
from tests.dbasserts import assert_sql_result


def test_demo_has_preloaded_languages(app_context):
    """
    When users get the initial demo, it has English, French, etc,
    pre-defined.
    """
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    sql = """
    select LgName
    from languages
    where LgName in ('English', 'French')
    order by LgName
    """
    assert_sql_result(sql, ["English", "French"], "sanity check loaded")


def test_new_language_has_sane_defaults():
    """
    Only validates the call to __init__.  Sqlalchemy mappings aren't used during the constuctor.
    """
    lang = Language()
    assert lang.character_substitutions == "´='|`='|’='|‘='|...=…|..=‥"
    assert lang.regexp_split_sentences == ".!?"
    assert lang.exceptions_split_sentences == "Mr.|Mrs.|Dr.|[A-Z].|Vd.|Vds."
    assert lang.word_characters == "a-zA-ZÀ-ÖØ-öø-ȳáéíóúÁÉÍÓÚñÑ"
    assert lang.right_to_left is False
    assert lang.show_romanization is False
    assert lang.parser_type == "spacedel"


def test_can_find_lang_by_name(app_context):
    """
    Returns lang if found, or None
    """
    lang = Language()
    lang.name = "English"
    db.session.add(lang)
    db.session.commit()
    repo = LanguageRepository(db.session)
    e = repo.find_by_name("English")
    assert e.name == "English", "case match"

    e_lc = repo.find_by_name("english")
    assert e_lc.name == "English", "case-insensitive"

    nf = repo.find_by_name("notfound")
    assert nf is None, "not found"


def test_language_word_char_regex_returns_python_compatible_regex(app_context):
    """
    Old Lute v2 ran in php, so the language word chars regex
    could look like this:

    x{0600}-x{06FF}x{FE70}-x{FEFC}  (where x = backslash-x)

    This needs to be converted to the python equivalent, e.g.

    u0600-u06FFuFE70-uFEFC  (where u = backslash-u)
    """
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    repo = LanguageRepository(db.session)
    a = repo.find_by_name("Arabic")
    # pylint: disable=line-too-long
    assert (
        a.word_characters
        == r"\u0600-\u0608\u060B\u060E-\u061A\u061C\u0620-\u0669\u066E-\u06D3\u06D5-\u06FF\uFE70-\uFEFC"
    )


def test_lang_to_dict_from_dict_returns_same_thing(app_context):
    """
    Lang can be exported to yaml and imported from yaml.
    A dictionary is used as the intermediary form, so the
    same language should return the same data.
    """
    demosvc = DemoService(db.session)
    demosvc.load_demo_data()
    repo = LanguageRepository(db.session)
    e = repo.find_by_name("English")
    e_dict = e.to_dict()
    e_from_dict = Language.from_dict(e_dict)
    e_back_to_dict = e_from_dict.to_dict()
    assert e_dict == e_back_to_dict, "Same thing returned"



"""
/term routes.
"""

import os
import csv
import json
from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    current_app,
    send_file,
    flash,
)
from lute.models.language import Language
from lute.models.term import Status
from lute.models.repositories import (
    LanguageRepository,
    TermRepository,
    UserSettingRepository,
)
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.term.datatables import get_data_tables_list
from lute.term.model import Repository, Term
from lute.term.service import (
    Service as TermService,
    TermServiceException,
    BulkTermUpdateData,
)
from lute.db import db
from lute.term.forms import TermForm
import lute.utils.formutils

bp = Blueprint("term", __name__, url_prefix="/term")


@bp.route("/index", defaults={"search": None}, methods=["GET"])
@bp.route("/index/<search>", methods=["GET"])
def index(search):
    "Index page."
    repo = TermRepository(db.session)
    repo.delete_empty_images()
    languages = db.session.query(Language).order_by(Language.name).all()
    langopts = [(lang.id, lang.name) for lang in languages]
    langopts = [(0, "(all)")] + langopts
    statuses = [s for s in db.session.query(Status).all() if s.id != Status.UNKNOWN]
    r = Repository(db.session)
    return render_template(
        "term/index.html",
        initial_search=search,
        language_options=langopts,
        statuses=statuses,
        tags=r.get_term_tags(),
        in_term_index_listing=True,
    )


def _load_term_custom_filters(request_form, parameters):
    "Manually add filters that the DataTablesFlaskParamParser doesn't know about."
    filter_param_names = [
        "filtLanguage",
        "filtParentsOnly",
        "filtAgeMin",
        "filtAgeMax",
        "filtStatusMin",
        "filtStatusMax",
        "filtIncludeIgnored",
    ]
    request_params = request_form.to_dict(flat=True)
    for p in filter_param_names:
        parameters[p] = request_params.get(p)


@bp.route("/datatables", methods=["POST"])
def datatables_active_source():
    "Datatables data for terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    data = get_data_tables_list(parameters, db.session)
    return jsonify(data)


def get_bulk_update_from_form(form):
    "Load the BulkTermUpdateData from the _bulk_edit_form_fields.html form."
    bud = BulkTermUpdateData()
    term_ids = form.get("term_ids").strip()
    if term_ids == "":
        return bud
    bud.term_ids = [int(tid.strip()) for tid in term_ids.split(",")]

    bud.lowercase_terms = form.get("lowercase_terms", "off") == "on"
    bud.remove_parents = form.get("remove_parents", "off") == "on"
    pdata = []
    if form.get("parent", "") != "":
        pdata = json.loads(form.get("parent"))
    if len(pdata) == 1:
        pdata = pdata[0]
        bud.parent_text = pdata.get("value")
        if "id" in pdata:
            bud.parent_id = int(pdata.get("id"))

    bud.change_status = form.get("change_status", "off") == "on"
    if "status" in form:
        bud.status_value = int(form.get("status"))

    def _get_tags(form_field_name):
        if form.get(form_field_name, "") == "":
            return []
        return [td["value"] for td in json.loads(form.get(form_field_name))]

    bud.add_tags = _get_tags("add_tags")
    bud.remove_tags = _get_tags("remove_tags")

    return bud


@bp.route("/bulk_edit_from_index", methods=["POST"])
def bulk_edit_from_index():
    "Edit from the term index listing."
    bud = get_bulk_update_from_form(request.form)
    svc = TermService(db.session)
    try:
        svc.apply_bulk_updates(bud)
    except TermServiceException as ex:
        flash(f"Error: {str(ex)}", "notice")
    return redirect("/term/index", 302)


@bp.route("/bulk_edit_from_reading_pane", methods=["POST"])
def bulk_edit_from_reading_pane():
    "Reading pane updates requires special redirect."
    bud = get_bulk_update_from_form(request.form)
    svc = TermService(db.session)
    try:
        svc.apply_bulk_updates(bud)
    except TermServiceException as ex:
        flash(f"Error: {str(ex)}", "notice")
        return redirect("/read/term_bulk_edit_form", 302)
    return render_template("/read/updated.html", term_text=None)


@bp.route("/export_terms", methods=["POST"])
def export_terms():
    "Generate export file of terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    parameters["length"] = 1000000
    outfile = os.path.join(current_app.env_config.temppath, "export_terms.csv")
    data = get_data_tables_list(parameters, db.session)
    term_data = data["data"]

    # Term data is an array of dicts, with the sql field name as dict
    # keys.  These need to be mapped to headings.
    heading_to_fieldname = {
        "term": "WoText",
        "parent": "ParentText",
        "translation": "WoTranslation",
        "language": "LgName",
        "tags": "TagList",
        "added": "WoCreated",
        "status": "StID",
        "link_status": "SyncStatus",
        "pronunciation": "WoRomanization",
    }

    headings = heading_to_fieldname.keys()
    output_data = [
        [r[heading_to_fieldname[fieldname]] for fieldname in headings]
        for r in term_data
    ]
    with open(outfile, "w", encoding="utf-8", newline="") as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(headings)
        csv_writer.writerows(output_data)

    return send_file(outfile, as_attachment=True, download_name="Terms.csv")


def handle_term_form(
    term,
    repo,
    session,
    form_template_name,
    return_on_success,
    embedded_in_reading_frame=False,
):  # pylint: disable=too-many-arguments,too-many-positional-arguments
    """
    Handle a form post.

    This is used here and in read.routes -- read.routes.term_form
    lives in an iframe in the reading frames and returns a different
    template on success.
    """
    form = TermForm(obj=term, session=session)

    # Flash messages get added on things like term imports.
    # The user opening the form is treated as an acknowledgement.
    term.flash_message = None

    form.language_id.choices = lute.utils.formutils.language_choices(session)

    if form.validate_on_submit():
        form.populate_obj(term)
        repo.add(term)
        repo.commit()
        return return_on_success

    # Note: on validation, form.duplicated_term may be set.
    # See DUPLICATE_TERM_CHECK comments in other files.

    hide_pronunciation = False
    language_repo = LanguageRepository(session)
    term_language = language_repo.find(
        term.language_id or -1
    )  # -1 hack for no lang set.
    if term_language is not None:
        hide_pronunciation = not term_language.show_romanization

    # Set the language dropdown to the user's current_language_id IF APPLICABLE.
    if embedded_in_reading_frame or term_language is not None:
        # Do nothing.  The language dropdown is not shown, or the term already
        # has a language assigned, and we shouldn't change it.
        pass
    else:
        # The language select control is shown and this is a new term,
        # so use the default value.
        us_repo = UserSettingRepository(db.session)
        current_language_id = int(us_repo.get_value("current_language_id"))
        form.language_id.data = current_language_id

    return render_template(
        form_template_name,
        form=form,
        term=term,
        duplicated_term=form.duplicated_term,
        language_dicts=language_repo.all_dictionaries(),
        hide_pronunciation=hide_pronunciation,
        tags=repo.get_term_tags(),
        embedded_in_reading_frame=embedded_in_reading_frame,
    )


def _handle_form(term, repo, redirect_to="/term/index"):
    """
    Handle the form post, redirecting to specified url.
    """
    return handle_term_form(
        term, repo, db.session, "/term/form.html", redirect(redirect_to, 302)
    )


@bp.route("/edit/<int:termid>", methods=["GET", "POST"])
def edit(termid):
    """
    Edit a term.
    """
    repo = Repository(db.session)
    term = repo.load(termid)
    if term.status == 0:
        term.status = 1
    return _handle_form(term, repo)


@bp.route("/editbytext/<int:langid>/<text>", methods=["GET", "POST"])
def edit_by_text(langid, text):
    """
    Edit a term.
    """
    repo = Repository(db.session)
    term = repo.find_or_new(langid, text)
    if term.status == 0:
        term.status = 1
    return _handle_form(term, repo)


@bp.route("/new", methods=["GET", "POST"])
def new():
    """
    Create a term.
    """
    repo = Repository(db.session)
    term = Term()
    return _handle_form(term, repo, "/term/new")


@bp.route("/search/<text>/<int:langid>", methods=["GET"])
def search_by_text_in_language(text, langid):
    "JSON data for parent data."
    if text.strip() == "" or langid == 0:
        return []
    repo = Repository(db.session)
    matches = repo.find_matches(langid, text)

    def _make_entry(t):
        return {
            "id": t.id,
            "text": t.text,
            "translation": t.translation,
            "status": t.status,
        }

    result = [_make_entry(t) for t in matches]
    return jsonify(result)


@bp.route("/sentences/<int:langid>/<text>", methods=["GET"])
def sentences(langid, text):
    "Get sentences for terms."
    repo = Repository(db.session)
    # Use find_or_new(): if the user clicks on a parent tag
    # in the term form, and the parent does not exist yet, then
    # we're creating a new term.
    t = repo.find_or_new(langid, text)
    refs = repo.find_references(t)

    # Transform data for output, to
    # { "term": [refs], "children": [refs], "parent1": [refs], "parent2" ... }
    refdata = [(f'"{text}"', refs["term"]), (f'"{text}" child terms', refs["children"])]
    for p in refs["parents"]:
        refdata.append((f"\"{p['term']}\"", p["refs"]))

    refcount = sum(len(ref[1]) for ref in refdata)
    return render_template(
        "/term/sentences.html",
        text=text,
        no_references=(refcount == 0),
        references=refdata,
    )


@bp.route("/bulk_update_status", methods=["POST"])
def bulk_update_status():
    """
    Update the statuses.

    json:
    {
      updates: [ { new_status: 1, termids: [ 42, ] }, ... }, ]
    }
    """
    repo = Repository(db.session)

    data = request.get_json()
    updates = data.get("updates")

    for u in updates:
        new_status = int(u.get("new_status"))
        termids = u.get("termids")
        for tidstring in termids:
            term = repo.load(int(tidstring))
            term.status = new_status
            repo.add(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/bulk_delete", methods=["POST"])
def bulk_delete():
    "Delete terms."
    data = request.get_json()
    termids = data.get("wordids")
    repo = Repository(db.session)
    for tid in termids:
        term = repo.load(int(tid))
        repo.delete(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/delete/<int:termid>", methods=["POST"])
def delete(termid):
    """
    Delete a term.
    """
    repo = Repository(db.session)
    term = repo.load(termid)
    repo.delete(term)
    repo.commit()
    return redirect("/term/index", 302)



"""
Data cleanup routines.

Sometimes required as data management changes.
These cleanup routines will be called by the app_factory.
"""

from sqlalchemy import text as sqltext
from lute.models.language import Language
from lute.models.book import Text, Sentence


class ProgressReporter:
    "Report progress for to an output function."

    def __init__(self, total_count, output_func, report_every=100):
        "Setup counters."
        self.current = 0
        self.last_output = 0
        self.total_count = total_count
        self.report_every = report_every
        self.output_func = output_func

    def increment(self):
        "Increment counter, and if past threshold, output."
        if self.total_count == 0:
            return
        self.current += 1
        if self.current - self.last_output < self.report_every:
            return
        self.output_func(f"  {self.current} of {self.total_count}")
        self.last_output = self.current


def _set_texts_word_count(session, output_function):
    """
    texts.TxWordCount should be set for all texts.

    Fixing a design error: the counts should have been stored here,
    instead of only in books.BkWordCount.

    Ref https://github.com/jzohrab/lute-v3/issues/95
    """
    calc_counts = session.query(Text).filter(Text.word_count.is_(None)).all()

    # Don't recalc with invalid parsers!!!!
    recalc = [t for t in calc_counts if t.book.language.is_supported]

    if len(recalc) == 0:
        # Nothing to calculate, quit.
        return

    output_function(f"Fixing word counts for {len(recalc)} Texts.")
    pr = ProgressReporter(len(recalc), output_function)
    for t in recalc:
        pr.increment()
        pt = t.book.language.get_parsed_tokens(t.text)
        words = [w for w in pt if w.is_word]
        t.word_count = len(words)
        session.add(t)
    session.commit()
    output_function("Done.")


def _load_sentence_textlc(session, output_function):
    """
    sentences.SeTextLC was added after deployment, need to load it
    to fix issue 531.  ref https://github.com/LuteOrg/lute-v3/issues/531

    Only update sentences where the language is supported.  e.g. the
    user may have installed mecab, done some japanese, and then
    uninstalled mecab: the data will be hidden, but it's still
    present, and the sentences cannot be updated as the parser can't
    be loaded.
    """

    supported_langs = {
        lang.id: lang for lang in session.query(Language).all() if lang.is_supported
    }
    langids = [f"{k}" for k in list(supported_langs.keys())]
    if len(langids) == 0:
        langids = ["-999"]  # dummy to ensure good base sql

    base_sql = f"""
    select SeID, BkLgID
    from sentences
    inner join texts on SeTxID = TxID
    inner join books on BkID = TxBkID
    where BkLgID in ({','.join(langids)})
    and SeTextLC is null
    """

    count = session.execute(sqltext(f"select count(*) from ({base_sql}) src")).scalar()
    if count == 0:
        # Do nothing, don't print messages."
        return

    def _get_next_batch(batch_size):
        # Query for up to 1000 Sentence objects where textlc_content is None
        sql = f"{base_sql} limit {batch_size}"
        recs = session.execute(sqltext(sql)).all()
        seids = [int(rec[0]) for rec in recs]
        if len(seids) == 0:
            return []

        sentences = session.query(Sentence).filter(Sentence.id.in_(seids)).all()
        se_map = {se.id: se for se in sentences}
        return [
            {"sentence": se_map[int(rec[0])], "langid": int(rec[1])} for rec in recs
        ]

    # Guard against infinite loop.
    last_batch_ids = []

    output_function(f"Updating data for {count} sentences.")
    batch_size = 1000
    pr = ProgressReporter(count, output_function, report_every=batch_size)
    batch = _get_next_batch(batch_size)
    while len(batch) > 0:
        curr_batch_ids = [se_langid["sentence"].id for se_langid in batch]
        if last_batch_ids == curr_batch_ids:
            raise RuntimeError("Sentences not getting updated correctly.")

        for se_langid in batch:
            pr.increment()
            sentence = se_langid["sentence"]
            lang = supported_langs[se_langid["langid"]]
            if lang is None:
                raise RuntimeError(f"Logic err: Missing langid={se_langid['langid']}")
            sentence.set_lowercase_text(lang.parser)
            session.add(sentence)
        session.commit()

        last_batch_ids = curr_batch_ids
        batch = _get_next_batch(batch_size)

    session.commit()
    output_function("Done.")


def clean_data(session, output_function):
    "Clean all data as required, sending messages to output_function."
    _set_texts_word_count(session, output_function)
    _load_sentence_textlc(session, output_function)



"""
TermTag tests.
"""

import pytest

from lute.models.term import TermTag
from lute.models.repositories import TermTagRepository
from lute.db import db
from tests.dbasserts import assert_sql_result


@pytest.fixture(name="_hola_tag")
def fixture_hola_tag(app_context):
    "A dummy tag for tests."
    t = TermTag("Hola", "Hola comment")
    db.session.add(t)
    db.session.commit()


def test_save(_hola_tag, app_context):
    "Save saves!"
    sql = "select TgID, TgText, TgComment from tags"
    expected = ["1; Hola; Hola comment"]
    assert_sql_result(sql, expected)


def test_new_dup_tag_text_fails(_hola_tag, app_context):
    "Smoke test of db integrity."
    t2 = TermTag("Hola", "dup")
    db.session.add(t2)
    with pytest.raises(Exception):
        db.session.commit()


def test_find_by_text(_hola_tag, app_context):
    "Find by text returns match."
    repo = TermTagRepository(db.session)
    retrieved = repo.find_by_text("Hola")
    assert retrieved is not None
    assert retrieved.text == "Hola"


def test_find_by_text_returns_null_if_not_exact_match(_hola_tag, app_context):
    "Find returns null if no match."
    repo = TermTagRepository(db.session)
    assert repo.find_by_text("unknown") is None
    assert repo.find_by_text("hola") is None


def test_find_or_create_by_text_returns_new_if_no_match(_hola_tag, app_context):
    "Return new."
    repo = TermTagRepository(db.session)
    assert repo.find_by_text("unknown") is None
    t = repo.find_or_create_by_text("unknown")
    assert t.text == "unknown", "new tag created"



"""
Render service tests.
"""

from lute.parse.base import ParsedToken
from lute.read.render.service import Service
from lute.db import db
from lute.models.term import Term

from tests.utils import add_terms, make_text, assert_rendered_text_equals
from tests.dbasserts import assert_sql_result


def _run_scenario(language, content, expected_found, msg=""):
    """
    Given some pre-saved terms in language,
    find_all method returns the expected_found terms that
    exist in the content string.
    """
    service = Service(db.session)
    found_terms = service.find_all_Terms_in_string(content, language)
    assert len(found_terms) == len(expected_found), "found count, " + msg
    zws = "\u200B"  # zero-width space
    found_terms = [t.text.replace(zws, "") for t in found_terms]
    assert found_terms is not None, msg
    assert expected_found is not None, msg
    found_terms.sort()
    expected_found.sort()
    assert found_terms == expected_found, msg


def test_smoke_tests(english, app_context):
    "Check bounds, ensure no false matches, etc."
    add_terms(english, ["a", "at", "xyz"])

    _run_scenario(english, "attack cat", [], "no matches, not standalone")
    _run_scenario(english, "at", ["at"], "a doesn't match, not standalone")
    _run_scenario(english, "A", ["a"], "case ignored")
    _run_scenario(english, "AT A", ["a", "at"], "case, order ignored")
    _run_scenario(english, "aatt", [], "no match")
    _run_scenario(english, "Xyz", ["xyz"], "case ignored 2")
    _run_scenario(english, "XyZ", ["xyz"], "case ignored 3")
    _run_scenario(english, "      A     at    x", ["a", "at"], "spaces ignored")

    _run_scenario(english, "a dog here", ["a"], "bounds check, found at start")
    _run_scenario(english, "dog a here", ["a"], "bounds check, found at start")
    _run_scenario(english, "dog here a", ["a"], "bounds check, found at end")
    _run_scenario(english, "a a a a a a a", ["a"], "return once only")

    add_terms(english, ["ab xy"])
    _run_scenario(english, "ab xy", ["ab xy"], "with space")
    _run_scenario(english, "cab xy", [], "extra at start")
    _run_scenario(english, "cab xyq", [], "no match, not the same")
    _run_scenario(english, "ab xyq", [], "extra stuff at end")


def test_spanish_find_all_in_string(spanish, app_context):
    "Given various pre-saved terms, find_all returns those in the string."
    add_terms(spanish, ["perro", "gato", "un gato"])

    _run_scenario(spanish, "Hola tengo un gato", ["gato", "un gato"])
    _run_scenario(spanish, "gato gato gato", ["gato"])
    _run_scenario(spanish, "No tengo UN PERRO", ["perro"])
    _run_scenario(spanish, "Hola tengo un    gato", ["gato", "un gato"])
    _run_scenario(spanish, "No tengo nada", [])

    add_terms(spanish, ["échalo", "ábrela"])

    _run_scenario(spanish, '"Échalo", me dijo.', ["échalo"])
    _run_scenario(spanish, "gato ábrela Ábrela", ["gato", "ábrela"])


def test_english_find_all_in_string(english, app_context):
    "Can find a term with an apostrophe in string."
    add_terms(english, ["the cat's pyjamas"])

    _run_scenario(english, "This is the cat's pyjamas.", ["the cat's pyjamas"])


def test_turkish_find_all_in_string(turkish, app_context):
    "Finds terms, handling case conversion."
    add_terms(turkish, ["ışık", "için"])

    _run_scenario(turkish, "Işık İçin.", ["ışık", "için"])


def test_smoke_get_paragraphs(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])
    perro = Term(spanish, "perro")
    perro.status = 0
    db.session.add(perro)

    content = "Tengo un gato. Hay un perro.\nTengo un perro."
    t = make_text("Hola", content, spanish)
    db.session.add(t)
    db.session.commit()

    sql = "select WoText from words order by WoText"
    assert_sql_result(sql, ["perro", "tengo/ /un", "un/ /gato"], "initial")

    ParsedToken.reset_counters()
    service = Service(db.session)
    paras = service.get_paragraphs(t.text, t.book.language)
    assert len(paras) == 2

    def stringize(t):
        zws = chr(0x200B)
        parts = [
            f"[{t.display_text.replace(zws, '/')}(",
            f"{t.paragraph_number}.{t.sentence_number}",
            ")]",
        ]
        return "".join(parts)

    sentences = [item for sublist in paras for item in sublist]
    actual = []
    for sentence in sentences:
        actual.append("".join(map(stringize, sentence)))

    expected = [
        "[Tengo/ /un(0.0)][ /gato(0.0)][. (0.0)]",
        "[Hay(0.1)][ (0.1)][un(0.1)][ (0.1)][perro(0.1)][.(0.1)]",
        "[Tengo/ /un(1.3)][ (1.3)][perro(1.3)][.(1.3)]",
    ]
    assert actual == expected

    assert_sql_result(sql, ["perro", "tengo/ /un", "un/ /gato"], "No new terms")


def test_smoke_rendered(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])
    content = ["Tengo un gato. Hay un perro.", "Tengo un perro."]
    text = make_text("Hola", "\n".join(content), spanish)
    db.session.add(text)
    db.session.commit()

    expected = ["Tengo un(1)/ gato(1)/. /Hay/ /un/ /perro/.", "Tengo un(1)/ /perro/."]
    assert_rendered_text_equals(text, expected)


def test_rendered_leaves_blank_lines(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])
    content = ["Tengo un gato. Hay un perro.", "", "Tengo un perro."]
    text = make_text("Hola", "\n".join(content), spanish)
    db.session.add(text)
    db.session.commit()

    expected = [
        "Tengo un(1)/ gato(1)/. /Hay/ /un/ /perro/.",
        "",
        "Tengo un(1)/ /perro/.",
    ]
    assert_rendered_text_equals(text, expected)



"""
Language entity.
"""

import re
from lute.db import db
from lute.parse.registry import get_parser, is_supported


class LanguageDictionary(db.Model):
    """
    Language dictionary.
    """

    __tablename__ = "languagedicts"

    id = db.Column("LdID", db.SmallInteger, primary_key=True)
    language_id = db.Column(
        "LdLgID", db.Integer, db.ForeignKey("languages.LgID"), nullable=False
    )
    language = db.relationship("Language", back_populates="dictionaries")
    usefor = db.Column("LdUseFor", db.String(20), nullable=False)
    dicttype = db.Column("LdType", db.String(20), nullable=False)
    dicturi = db.Column("LdDictURI", db.String(200), nullable=False)
    is_active = db.Column("LdIsActive", db.Boolean, default=True)
    sort_order = db.Column("LdSortOrder", db.SmallInteger, nullable=False)

    # HACK: pre-pend '*' to URLs that need to open a new window.
    # This is a relic of the original code, and should be changed.
    # TODO remove-asterisk-hack: remove * from URL start.
    def make_uri(self):
        "Hack add asterisk."
        prepend = "*" if self.dicttype == "popuphtml" else ""
        return f"{prepend}{self.dicturi}"


class Language(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Language entity.
    """

    __tablename__ = "languages"

    id = db.Column("LgID", db.SmallInteger, primary_key=True)
    name = db.Column("LgName", db.String(40))

    dictionaries = db.relationship(
        "LanguageDictionary",
        back_populates="language",
        order_by="LanguageDictionary.sort_order",
        lazy="subquery",
        cascade="all, delete-orphan",
    )

    character_substitutions = db.Column("LgCharacterSubstitutions", db.String(500))
    regexp_split_sentences = db.Column("LgRegexpSplitSentences", db.String(500))
    exceptions_split_sentences = db.Column("LgExceptionsSplitSentences", db.String(500))
    _word_characters = db.Column("LgRegexpWordCharacters", db.String(500))
    right_to_left = db.Column("LgRightToLeft", db.Boolean)
    show_romanization = db.Column("LgShowRomanization", db.Boolean)
    parser_type = db.Column("LgParserType", db.String(20))

    def __init__(self):
        self.character_substitutions = "´='|`='|’='|‘='|...=…|..=‥"
        self.regexp_split_sentences = ".!?"
        self.exceptions_split_sentences = "Mr.|Mrs.|Dr.|[A-Z].|Vd.|Vds."
        self.word_characters = "a-zA-ZÀ-ÖØ-öø-ȳáéíóúÁÉÍÓÚñÑ"
        self.right_to_left = False
        self.show_romanization = False
        self.parser_type = "spacedel"
        self.dictionaries = []

    def __repr__(self):
        return f"<Language {self.id} '{self.name}'>"

    def _get_python_regex_pattern(self, s):
        """
        Old Lute v2 ran in php, so the language word chars regex
        could look like this:

        x{0600}-x{06FF}x{FE70}-x{FEFC}  (where x = backslash-x)

        This needs to be converted to the python equivalent, e.g.

        u0600-u06FFuFE70-uFEFC  (where u = backslash-u)
        """

        def convert_match(match):
            # Convert backslash-x{XXXX} to backslash-uXXXX
            hex_value = match.group(1)
            return f"\\u{hex_value}"

        ret = re.sub(r"\\x{([0-9A-Fa-f]+)}", convert_match, s)
        return ret

    @property
    def word_characters(self):
        return self._get_python_regex_pattern(self._word_characters)

    @word_characters.setter
    def word_characters(self, s):
        self._word_characters = self._get_python_regex_pattern(s)

    def active_dict_uris(self, use_for):
        "Get sorted uris for active dicts of correct type."
        actives = [d for d in self.dictionaries if d.is_active and d.usefor == use_for]
        sorted_actives = sorted(actives, key=lambda x: x.sort_order)
        return [d.make_uri() for d in sorted_actives]

    @property
    def sentence_dict_uris(self):
        return self.active_dict_uris("sentences")

    @property
    def parser(self):
        "Note: this throws if the parser is not supported!!!"
        return get_parser(self.parser_type)

    @property
    def is_supported(self):
        "True if the language's parser is supported."
        return is_supported(self.parser_type)

    def get_parsed_tokens(self, s):
        return self.parser.get_parsed_tokens(s, self)

    def get_lowercase(self, s) -> str:
        return self.parser.get_lowercase(s)

    def to_dict(self):
        "Return dictionary of data, for serialization."
        ret = {}
        ret["name"] = self.name
        ret["dictionaries"] = []
        for d in self.dictionaries:
            dd = {}
            dd["for"] = d.usefor
            dd["type"] = d.dicttype.replace("html", "")
            dd["url"] = d.dicturi
            dd["active"] = d.is_active
            ret["dictionaries"].append(dd)
        ret["show_romanization"] = self.show_romanization
        ret["right_to_left"] = self.right_to_left
        ret["parser_type"] = self.parser_type
        ret["character_substitutions"] = self.character_substitutions
        ret["split_sentences"] = self.regexp_split_sentences
        ret["split_sentence_exceptions"] = self.exceptions_split_sentences
        ret["word_chars"] = self.word_characters
        return ret

    @staticmethod
    def from_dict(d):
        "Create new Language from dictionary d."

        lang = Language()

        def load(key, method):
            if key in d:
                val = d[key]
                # Handle boolean values
                if isinstance(val, str):
                    temp = val.lower()
                    if temp == "true":
                        val = True
                    elif temp == "false":
                        val = False
                setattr(lang, method, val)

        # Define mappings for fields
        mappings = {
            "name": "name",
            "show_romanization": "show_romanization",
            "right_to_left": "right_to_left",
            "parser_type": "parser_type",
            "character_substitutions": "character_substitutions",
            "split_sentences": "regexp_split_sentences",
            "split_sentence_exceptions": "exceptions_split_sentences",
            "word_chars": "word_characters",
        }

        for key in d.keys():
            funcname = mappings.get(key, "")
            if funcname:
                load(key, funcname)

        ld_sort = 1
        for ld_data in d["dictionaries"]:
            dtype = ld_data["type"]
            if dtype == "embedded":
                dtype = "embeddedhtml"
            elif dtype == "popup":
                dtype = "popuphtml"
            else:
                raise ValueError(f"Invalid dictionary type {dtype}")

            ld = LanguageDictionary()
            # ld.language = lang -- if you do this, the dict is added twice.
            ld.usefor = ld_data["for"]
            ld.dicttype = dtype
            ld.dicturi = ld_data["url"]
            ld.is_active = ld_data.get("active", True)

            ld.sort_order = ld_sort
            ld_sort += 1
            lang.dictionaries.append(ld)

        return lang



"""
Term popup data tests.
"""

import pytest
from lute.models.term import Term, TermTag, Status
from lute.models.repositories import UserSettingRepository
from lute.read.service import Service
from lute.db import db


@pytest.fixture(name="service")
def fixture_svc(app_context):
    "Service"
    return Service(db.session)


def test_popup_data_is_none_if_no_data(spanish, app_context, service):
    "Return None if no popup."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d is None, "No data, no popup"

    t.translation = "hello"
    d = service.get_popup_data(t.id)
    assert d is not None, "Have data, popup"


def test_popup_data_shown_if_have_tag(spanish, app_context, service):
    "Return None if no popup."
    t = Term(spanish, "gato")
    t.add_term_tag(TermTag("animal"))
    db.session.add(t)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d is not None, "Have tag, show popup"
    assert d.tags == ["animal"]


def test_popup_shown_if_parent_exists_even_if_no_other_data(
    spanish, app_context, service
):
    "I always want to know if a parent has been set."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d is None, "No data, no popup"

    p = Term(spanish, "perro")
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d is not None, "Has parent, popup, even if no other data."


def test_popup_data_is_shown_if_have_data_regardless_of_status(
    spanish, app_context, service
):
    "Return None if no-popup statuses."
    t = Term(spanish, "gato")
    t.translation = "hello"
    for s in [1, Status.UNKNOWN, Status.IGNORED]:
        t.status = s
        db.session.add(t)
        db.session.commit()
        d = service.get_popup_data(t.id)
        assert d is not None, f"Have data for status {s}"


def test_term_with_no_parents(spanish, app_context, service):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    t.translation = "cat"
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.parents == [], "no parents"
    assert d.parents_text == "", "no parents"


def test_parent_not_shown_if_has_no_useful_data(spanish, app_context, service):
    "No need for dup data"
    t = Term(spanish, "gato")
    p = Term(spanish, "perro")
    p.translation = "cat"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "cat", "trans promoted"
    assert len(d.parents) == 0, "no data"

    p.romanization = "hello"
    db.session.add(p)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert len(d.parents) == 1, "some data"


def test_images_combined_in_popup(spanish, app_context, service):
    "combine images as needed."
    t = Term(spanish, "gato")
    t.set_current_image("gato.jpg")
    p = Term(spanish, "perro")
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.popup_image_data == {"gato.jpg": "gato"}

    p.set_current_image("perro.jpg")
    db.session.add(p)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d.popup_image_data == {"gato.jpg": "gato", "perro.jpg": "perro"}

    p.set_current_image("gato.jpg")
    db.session.add(p)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert d.popup_image_data == {"gato.jpg": "gato, perro"}


def test_single_parent_translation_can_be_promoted_to_term_if_term_translation_blank(
    spanish, app_context, service
):
    "No need for dup data"
    t = Term(spanish, "gato")
    p = Term(spanish, "perro")
    p.translation = "cat"
    p.romanization = "hello"  # need extra data
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "cat", "trans promoted"
    assert d.parents[0].translation == "", "moved up"

    repo = UserSettingRepository(db.session)
    repo.set_value("term_popup_promote_parent_translation", False)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "", "trans not promoted"
    assert d.parents[0].translation == "cat", "translation left with parent"


def test_single_parent_translation_may_be_removed_if_same_as_child(
    spanish, app_context, service
):
    "No need for dup data"
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "cat"
    p.romanization = "hello"  # need extra data
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "cat", "trans promoted"
    assert d.parents[0].translation == "", "moved up"

    repo = UserSettingRepository(db.session)
    repo.set_value("term_popup_promote_parent_translation", False)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "cat", "trans left"
    assert d.parents[0].translation == "cat", "translation also left with parent"


def test_multiple_parents_translations_left_alone_even_if_blank(
    spanish, app_context, service
):
    "No need for dup data"
    t = Term(spanish, "gato")
    p = Term(spanish, "perro")
    p.translation = "cat"
    p2 = Term(spanish, "otro")
    p2.translation = "cat"
    t.parents.append(p)
    t.parents.append(p2)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.translation == "", "still blank"
    assert d.parents[0].translation == "cat", "not moved up"


def test_parent_data_returned(spanish, app_context, service):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    p = d.parents[0]
    assert p.term_text == "perro"
    assert p.romanization == ""
    assert p.translation == "kitty"
    assert p.tags == []
    assert d.parents_text == "perro", "one parent"


def test_parent_data_always_added_if_multiple_parents(spanish, app_context, service):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    p2 = Term(spanish, "hombre")
    p2.translation = "kitteh"
    t.parents.append(p)
    t.parents.append(p2)
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert len(d.parents) == 2, "extra parent data added"
    assert d.parents_text == "perro, hombre", "parents"


## Component term checks.


def make_terms(term_trans_pairs, spanish):
    "Make test data"
    for c in term_trans_pairs:
        ct = Term(spanish, c[0])
        ct.translation = c[1]
        db.session.add(ct)


def assert_components(d, expected, msg=""):
    "Check components."

    def _c_to_string(c):
        s = "; ".join([c.term_text, c.translation])
        zws = chr(0x200B)
        return s.replace(zws, "")

    actual = [_c_to_string(c) for c in d.components]
    assert actual == expected, msg


def test_single_term_not_included_in_own_components(spanish, app_context, service):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    t.translation = "cat"
    db.session.add(t)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert d.components == [], "no components"


def test_component_without_useful_data_not_returned(spanish, app_context, service):
    "Component word is returned."
    t = Term(spanish, "un gato")
    t.translation = "a cat"
    g = Term(spanish, "gato")
    db.session.add(t)
    db.session.add(g)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert len(d.components) == 0, "no component data"

    g.translation = "something"
    db.session.add(g)
    db.session.commit()
    d = service.get_popup_data(t.id)
    assert len(d.components) == 1, "have data"
    assert d.components[0].translation == "something", "check"


def test_component_word_with_translation_returned(spanish, app_context, service):
    "Component word is returned."
    t = Term(spanish, "un gato")
    t.translation = "a cat"
    db.session.add(t)
    make_terms([("gato", "cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["gato; cat"], "one component")


def test_nested_multiword_components_returned_depending_on_setting(
    spanish, app_context, service
):
    "Complete components are returned."
    t = Term(spanish, "un gato gordo")
    t.translation = "a fat cat"
    db.session.add(t)
    make_terms([("gato", "cat"), ("gat", "x"), ("un gato", "a cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["un gato; a cat", "gato; cat"], "components")

    repo = UserSettingRepository(db.session)
    repo.set_value("term_popup_show_components", False)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, [], "no components shown")


def test_multiword_components_returned_in_order_of_appearance(
    spanish, app_context, service
):
    "Complete components are returned."
    t = Term(spanish, "un gato gordo")
    t.translation = "a fat cat"
    db.session.add(t)
    make_terms(
        [
            ("gato", "cat"),
            ("gat", "x"),
            ("gato gordo", "fat"),
            ("un gato", "a cat"),
        ],
        spanish,
    )
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(
        d, ["un gato; a cat", "gato gordo; fat", "gato; cat"], "components"
    )


def test_components_only_returned_once(spanish, app_context, service):
    "Component not returned multiple times if present multiple times."
    t = Term(spanish, "un gato gordo gato")
    t.translation = "a cat fat cat"
    db.session.add(t)
    make_terms([("gato", "cat")], spanish)
    db.session.commit()

    d = service.get_popup_data(t.id)
    assert_components(d, ["gato; cat"], "components")



"""
Tests for managing the demo data.

Prior to https://github.com/LuteOrg/lute-v3/issues/534, the baseline
db had languages and stories pre-loaded.  This created a lot of db
file thrash whenever that data changed.

In the new setup, the baseline db only contains a flag, "LoadDemoData"
(true/false), which is initially set to True.  When the app first
starts up, if that flag is True, it loads the demo data, and sets
"LoadDemoData" to False, and "IsDemoData" to True.

If the LoadDemoData flag is set, the demo data is loaded from the
startup scripts (devstart and lute.main)

"""

from sqlalchemy import text
import pytest
from lute.db import db
from lute.db.demo import Service
import lute.parse.registry
from tests.dbasserts import assert_record_count_equals, assert_sql_result


# ========================================
# See notes at top of file re these tests.
# ========================================


@pytest.fixture(name="service")
def _service(app_context):
    return Service(db.session)


def test_new_db_doesnt_contain_anything(service):
    "New db created from the baseline has the demo flag set."
    assert service.should_load_demo_data() is True, "has LoadDemoData flag."
    assert service.contains_demo_data() is False, "no demo data."


def test_empty_db_not_loaded_if_load_flag_not_set(service):
    "Even if it's empty, nothing happens."
    service.remove_load_demo_flag()
    assert service.contains_demo_data() is False, "no demo data."
    assert_record_count_equals("select * from languages", 0, "empty")
    service.load_demo_data()
    assert service.contains_demo_data() is False, "no demo data."
    assert service.should_load_demo_data() is False, "still no reload."
    assert_record_count_equals("select * from languages", 0, "still empty")


def test_smoke_test_load_demo_works(service):
    "Wipe everything, but set the flag and then start."
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.contains_demo_data() is True, "demo loaded."
    assert service.tutorial_book_id() > 0, "Have tutorial"
    assert service.should_load_demo_data() is False, "loaded once, don't reload."


def test_load_not_run_if_data_exists_even_if_flag_is_set(service):
    "Just in case."
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.tutorial_book_id() > 0, "Have tutorial"
    assert service.should_load_demo_data() is False, "loaded once, don't reload."

    service.remove_flag()
    service.set_load_demo_flag()
    assert service.should_load_demo_data() is True, "should re-reload demo data."
    service.load_demo_data()  # if this works, it didn't throw :-P
    assert service.should_load_demo_data() is False, "already loaded once."


def test_removing_flag_means_not_demo(service):
    "Unsetting the flag means the db is not a demo."
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.contains_demo_data() is True, "demo loaded."
    service.remove_flag()
    assert service.contains_demo_data() is False, "not a demo now."


def test_wiping_db_clears_flag(service):
    "No longer a demo if the demo is wiped out!"
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.contains_demo_data() is True, "demo loaded."
    service.delete_demo_data()
    assert service.contains_demo_data() is False, "not a demo."


def test_wipe_db_only_works_if_flag_is_set(service):
    "Can only wipe a demo db!"
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.contains_demo_data() is True, "demo loaded."
    service.remove_flag()
    with pytest.raises(Exception):
        service.delete_demo_data()


def test_tutorial_id_returned_if_present(service):
    "Sanity check."
    assert service.should_load_demo_data() is True, "should reload demo data."
    service.load_demo_data()
    assert service.tutorial_book_id() > 0, "have tutorial"

    sql = 'update books set bktitle = "xxTutorial" where bktitle = "Tutorial"'
    db.session.execute(text(sql))
    db.session.commit()
    assert service.tutorial_book_id() is None, "no tutorial"

    sql = 'update books set bktitle = "Tutorial" where bktitle = "xxTutorial"'
    db.session.execute(text(sql))
    db.session.commit()
    assert service.tutorial_book_id() > 0, "have tutorial again"

    service.delete_demo_data()
    assert service.tutorial_book_id() is None, "no tutorial"


# Loading.


@pytest.mark.dbreset
def test_rebaseline(service):
    """
    This test is also used from "inv db.reset" in tasks.py
    (see .pytest.ini).
    """
    assert service.contains_demo_data() is False, "not a demo."
    assert_record_count_equals("languages", 0, "wiped out")

    # Wipe out all user settings!!!  When user installs and first
    # starts up, the user settings need to be loaded with values from
    # _their_ config and environment.
    sql = "delete from settings where StKeyType = 'user'"
    db.session.execute(text(sql))
    db.session.commit()

    service.set_load_demo_flag()

    sql = "select stkeytype, stkey, stvalue from settings"
    assert_sql_result(sql, ["system; LoadDemoData; 1"], "only this key is set.")


@pytest.fixture(name="_restore_japanese_parser")
def fixture_restore_mecab_support():
    """
    "Teardown" method to restore jp parser if it was removed.
    """
    k = "japanese"
    assert k in lute.parse.registry.__LUTE_PARSERS__, "have jp parser, sanity check"
    old_val = lute.parse.registry.__LUTE_PARSERS__[k]

    yield

    if k not in lute.parse.registry.__LUTE_PARSERS__:
        lute.parse.registry.__LUTE_PARSERS__[k] = old_val



"""
Parsing for space-delimited languages.

The parser uses some Language settings (e.g., word characters) to
perform the actual parsing.

Includes classes:

- SpaceDelimitedParser
- TurkishParser
"""

import functools
import re
import sys
import unicodedata

from typing import List

from lute.parse.base import ParsedToken, AbstractParser


class SpaceDelimitedParser(AbstractParser):
    """
    A general parser for space-delimited languages,
    such as English, French, Spanish ... etc.
    """

    @classmethod
    def name(cls):
        return "Space Delimited"

    @staticmethod
    @functools.lru_cache
    def compile_re_pattern(pattern: str, *args, **kwargs) -> re.Pattern:
        """Compile regular expression pattern, cache result for fast re-use."""
        return re.compile(pattern, *args, **kwargs)

    @staticmethod
    @functools.lru_cache
    def get_default_word_characters() -> str:
        """Return default value for lang.word_characters."""

        # Unicode categories reference: https://www.compart.com/en/unicode/category
        categories = set(["Cf", "Ll", "Lm", "Lo", "Lt", "Lu", "Mc", "Mn", "Sk"])

        # There are more than 130,000 characters across all these categories.
        # Expressing this a single character at a time, mostly using unicode
        # escape sequences like \u1234 or \U12345678, would require 1 megabyte.
        # Converting to ranges like \u1234-\u1256 requires only 10K.
        ranges = []
        current = None

        def add_current_to_ranges():
            def ucode(n):
                "Unicode point for integer."
                fstring = r"\u{:04x}" if n < 0x10000 else r"\U{:08x}"
                return (fstring).format(n)

            start_code = ucode(current[0])
            if current[0] == current[1]:
                range_string = start_code
            else:
                endcode = ucode(current[1])
                range_string = f"{start_code}-{endcode}"
            ranges.append(range_string)

        for i in range(1, sys.maxunicode):
            if unicodedata.category(chr(i)) not in categories:
                if current is not None:
                    add_current_to_ranges()
                    current = None
            elif current is None:
                # Starting a new range.
                current = [i, i]
            else:
                # Extending existing range.
                current[1] = i

        if current is not None:
            add_current_to_ranges()

        return "".join(ranges)

    @staticmethod
    @functools.lru_cache
    def get_default_regexp_split_sentences() -> str:
        """Return default value for lang.regexp_split_sentences."""

        # Construct pattern from Unicode ATerm and STerm categories.
        # See: https://www.unicode.org/Public/UNIDATA/auxiliary/SentenceBreakProperty.txt
        # and: https://unicode.org/reports/tr29/

        # Also include colon, since that is used to separate speakers
        # and their dialog, and is a reasonable dividing point for
        # sentence translations.

        return "".join(
            [
                re.escape(".!?:"),
                # ATerm entries (other than ".", covered above):
                r"\u2024\uFE52\uFF0E",
                # STerm entries (other than "!" and "?", covered above):
                r"\u0589",
                r"\u061D-\u061F\u06D4",
                r"\u0700-\u0702",
                r"\u07F9",
                r"\u0837\u0839\u083D\u083E",
                r"\u0964\u0965",
                r"\u104A\u104B",
                r"\u1362\u1367\u1368",
                r"\u166E",
                r"\u1735\u1736",
                r"\u17D4\u17D5",
                r"\u1803\u1809",
                r"\u1944\u1945",
                r"\u1AA8-\u1AAB",
                r"\u1B5A\u1B5B\u1B5E\u1B5F\u1B7D\u1B7E",
                r"\u1C3B\u1C3C",
                r"\u1C7E\u1C7F",
                r"\u203C\u203D\u2047-\u2049\u2E2E\u2E3C\u2E53\u2E54\u3002",
                r"\uA4FF",
                r"\uA60E\uA60F",
                r"\uA6F3\uA6F7",
                r"\uA876\uA877",
                r"\uA8CE\uA8CF",
                r"\uA92F",
                r"\uA9C8\uA9C9",
                r"\uAA5D\uAA5F",
                r"\uAAF0\uAAF1\uABEB",
                r"\uFE56\uFE57\uFF01\uFF1F\uFF61",
                r"\U00010A56\U00010A57",
                r"\U00010F55-\U00010F59",
                r"\U00010F86-\U00010F89",
                r"\U00011047\U00011048",
                r"\U000110BE-\U000110C1",
                r"\U00011141-\U00011143",
                r"\U000111C5\U000111C6\U000111CD\U000111DE\U000111DF",
                r"\U00011238\U00011239\U0001123B\U0001123C",
                r"\U000112A9",
                r"\U0001144B\U0001144C",
                r"\U000115C2\U000115C3\U000115C9-\U000115D7",
                r"\U00011641\U00011642",
                r"\U0001173C-\U0001173E",
                r"\U00011944\u00011946",
                r"\U00011A42\U00011A43",
                r"\U00011A9B\U00011A9C",
                r"\U00011C41\U00011C42",
                r"\U00011EF7\U00011EF8",
                r"\U00011F43\U00011F44",
                r"\U00016A6E\U00016A6F",
                r"\U00016AF5",
                r"\U00016B37\U00016B38\U00016B44",
                r"\U00016E98",
                r"\U0001BC9F",
                r"\U0001DA88",
            ]
        )

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        "Return parsed tokens."

        # Remove extra spaces.
        clean_text = re.sub(r" +", " ", text)

        # Remove zero-width spaces.
        clean_text = clean_text.replace(chr(0x200B), "")

        return self._parse_to_tokens(clean_text, language)

    def preg_match_capture(self, pattern, subject):
        """
        Return the matched text and their start positions in the subject.

        E.g. search for r'cat' in "there is a CAT and a Cat" returns:
        [['CAT', 11], ['Cat', 21]]
        """
        compiled = SpaceDelimitedParser.compile_re_pattern(pattern, flags=re.IGNORECASE)
        matches = compiled.finditer(subject)
        result = [[match.group(), match.start()] for match in matches]
        return result

    def _parse_to_tokens(self, text: str, lang):
        """
        Returns ParsedToken array for given language.
        """
        replacements = lang.character_substitutions.split("|")
        for replacement in replacements:
            fromto = replacement.strip().split("=")
            if len(fromto) >= 2:
                rfrom = fromto[0].strip()
                rto = fromto[1].strip()
                text = text.replace(rfrom, rto)

        text = text.replace("\r\n", "\n")
        text = text.replace("{", "[")
        text = text.replace("}", "]")

        tokens = []
        paras = text.split("\n")
        pcount = len(paras)
        for i, para in enumerate(paras):
            self.parse_para(para, lang, tokens)
            if i != (pcount - 1):
                tokens.append(ParsedToken("¶", False, True))

        return tokens

    def parse_para(self, text: str, lang, tokens: List[ParsedToken]):
        """
        Parse a string, appending the tokens to the list of tokens.
        """
        termchar = lang.word_characters.strip()
        if not termchar:
            termchar = SpaceDelimitedParser.get_default_word_characters()

        splitex = lang.exceptions_split_sentences.replace(".", "\\.")
        pattern = rf"({splitex}|[{termchar}]*)"
        if splitex.strip() == "":
            pattern = rf"([{termchar}]*)"

        m = self.preg_match_capture(pattern, text)
        wordtoks = list(filter(lambda t: t[0] != "", m))

        def add_non_words(s):
            """
            Add non-word token s to the list of tokens.  If s
            matches any of the split_sentence values, mark it as an
            end-of-sentence.
            """
            if not s:
                return
            splitchar = lang.regexp_split_sentences.strip()
            if not splitchar:
                splitchar = SpaceDelimitedParser.get_default_regexp_split_sentences()
            pattern = f"[{re.escape(splitchar)}]"
            has_eos = False
            if pattern != "[]":  # Should never happen, but ...
                allmatches = self.preg_match_capture(pattern, s)
                has_eos = len(allmatches) > 0
            tokens.append(ParsedToken(s, False, has_eos))

        # For each wordtok, add all non-words before the wordtok, and
        # then add the wordtok.
        pos = 0
        for wt in wordtoks:
            w = wt[0]
            wp = wt[1]
            s = text[pos:wp]
            add_non_words(s)
            tokens.append(ParsedToken(w, True, False))
            pos = wp + len(w)

        # Add anything left over.
        s = text[pos:]
        add_non_words(s)


class TurkishParser(SpaceDelimitedParser):
    "Parser to handle Turkish parsing fun."

    @classmethod
    def name(cls):
        return "Turkish"

    def get_lowercase(self, text):
        "Handle the funny turkish i variants."
        for caps, lower in {"İ": "i", "I": "ı"}.items():
            text = text.replace(caps, lower)
        return text.lower()



"""
Factory.

Methods: create_app.
"""

import os
import json
import platform
import traceback
from flask import (
    Flask,
    render_template,
    request,
    redirect,
    flash,
    current_app,
    make_response,
    send_from_directory,
    jsonify,
)
from sqlalchemy.event import listens_for
from sqlalchemy.pool import Pool

from lute.config.app_config import AppConfig
from lute.db import db
from lute.db.setup.main import setup_db
from lute.db.management import add_default_user_settings
from lute.db.data_cleanup import clean_data
from lute.backup.service import Service as BackupService
from lute.db.demo import Service as DemoService
import lute.utils.formutils

from lute.parse.registry import init_parser_plugins, supported_parsers

from lute.models.book import Book
from lute.models.language import Language
from lute.settings.current import (
    refresh_global_settings,
    current_settings,
    current_hotkeys,
)
from lute.models.repositories import UserSettingRepository
from lute.book.stats import Service as StatsService

from lute.book.routes import bp as book_bp
from lute.bookmarks.routes import bp as bookmarks_bp
from lute.language.routes import bp as language_bp
from lute.term.routes import bp as term_bp
from lute.termtag.routes import bp as termtag_bp
from lute.read.routes import bp as read_bp
from lute.bing.routes import bp as bing_bp
from lute.userimage.routes import bp as userimage_bp
from lute.useraudio.routes import bp as useraudio_bp
from lute.termimport.routes import bp as termimport_bp
from lute.backup.routes import bp as backup_bp
from lute.dev_api.routes import bp as dev_api_bp
from lute.settings.routes import bp as settings_bp
from lute.themes.routes import bp as themes_bp
from lute.stats.routes import bp as stats_bp
from lute.cli.commands import bp as cli_bp


def _setup_app_dir(dirname, readme_content):
    "Create one app directory."
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    readme = os.path.join(dirname, "README.md")
    if not os.path.exists(readme):
        with open(readme, "w", encoding="utf-8") as f:
            f.write(readme_content)


def _setup_app_dirs(app_config):
    """
    App needs the data dir, backups, and other directories.
    """
    dp = app_config.datapath
    required_dirs = [
        [dp, "Lute data folder."],
        [
            app_config.default_user_backup_path,
            "Default path for user backups, can be overridden in settings.",
        ],
        [
            app_config.system_backup_path,
            "Database backups created by Lute at app start, just in case.",
        ],
        [
            app_config.userimagespath,
            "User images.  Each subfolder is a language's ID.",
        ],
        [
            app_config.userthemespath,
            "User themes.  <theme_name>.css files for your personal themes.",
        ],
        [
            app_config.useraudiopath,
            "User audio.  Each file is a book's audio.",
        ],
        [
            app_config.temppath,
            "Temp directory for export file writes, to avoid permissions issues.",
        ],
    ]
    for rec in required_dirs:
        _setup_app_dir(rec[0], rec[1])


def _add_base_routes(app, app_config):
    """
    Add some basic routes.
    """

    @app.context_processor
    def inject_menu_bar_vars():
        """
        Inject backup settings into the all templates for the menu bar.
        """
        us_repo = UserSettingRepository(db.session)
        bs = us_repo.get_backup_settings()
        have_languages = len(db.session.query(Language).all()) > 0
        ret = {
            "have_languages": have_languages,
            "backup_enabled": bs.backup_enabled,
            "backup_directory": bs.backup_dir,
            "backup_last_display_date": bs.last_backup_display_date,
            "backup_time_since": bs.time_since_last_backup,
            "user_settings": json.dumps(current_settings),
            "user_hotkeys": json.dumps(current_hotkeys),
        }
        return ret

    @app.route("/")
    def index():
        demosvc = DemoService(db.session)
        is_production = not demosvc.contains_demo_data()
        us_repo = UserSettingRepository(db.session)
        bkp_settings = us_repo.get_backup_settings()

        have_books = len(db.session.query(Book).all()) > 0
        have_languages = len(db.session.query(Language).all()) > 0
        language_choices = lute.utils.formutils.language_choices(
            db.session, "(all languages)"
        )
        current_language_id = lute.utils.formutils.valid_current_language_id(db.session)

        bs = BackupService(db.session)
        should_run_auto_backup = bs.should_run_auto_backup(bkp_settings)
        # Only back up if we have books, otherwise the backup is
        # kicked off when the user empties the demo database.
        if is_production and have_books and should_run_auto_backup:
            return redirect("/backup/backup", 302)

        warning_msg = bs.backup_warning(bkp_settings)
        backup_show_warning = (
            bkp_settings.backup_warn
            and bkp_settings.backup_enabled
            and warning_msg != ""
        )

        demosvc = DemoService(db.session)
        response = make_response(
            render_template(
                "index.html",
                hide_homelink=True,
                dbname=app_config.dbname,
                datapath=app_config.datapath,
                tutorial_book_id=demosvc.tutorial_book_id(),
                have_books=have_books,
                have_languages=have_languages,
                language_choices=language_choices,
                current_language_id=current_language_id,
                is_production_data=is_production,
                backup_show_warning=backup_show_warning,
                backup_warning_msg=warning_msg,
            )
        )
        return response

    @app.route("/refresh_all_stats")
    def refresh_all_stats():
        books_to_update = db.session.query(Book).filter(Book.archived == 0).all()
        svc = StatsService(db.session)
        for book in books_to_update:
            svc.mark_stale(book)
        return redirect("/", 302)

    @app.route("/wipe_database")
    def wipe_db():
        demosvc = DemoService(db.session)
        if demosvc.contains_demo_data():
            demosvc.delete_demo_data()
            msg = """
            The database has been wiped clean.  Have fun! <br /><br />
            <i>(Lute has automatically enabled backups --
            change your <a href="/settings/index">Settings</a> as needed.)</i>
            """
            flash(msg)
        return redirect("/", 302)

    @app.route("/remove_demo_flag")
    def remove_demo():
        demosvc = DemoService(db.session)
        if demosvc.contains_demo_data():
            demosvc.remove_flag()
            msg = """
            Demo mode deactivated. Have fun! <br /><br />
            <i>(Lute has automatically enabled backups --
            change your <a href="/settings/index">Settings</a> as needed.)</i>
                        """
            flash(msg)
        return redirect("/", 302)

    @app.route("/version")
    def show_version():
        ac = current_app.env_config
        return render_template(
            "version.html",
            version=lute.__version__,
            datapath=ac.datapath,
            database=ac.dbfilename,
            is_docker=ac.is_docker,
        )

    @app.route("/info")
    def show_info():
        """
        Json return of some data.

        Used in lute.verify module for tests.

        This likely belongs in a different 'api' location,
        but leaving it here for now.
        """
        ret = {
            "version": lute.__version__,
            "datapath": current_app.config["DATAPATH"],
            "database": current_app.config["DATABASE"],
        }
        return jsonify(ret)

    @app.route("/static/js/never_cache/<path:filename>")
    def custom_js(filename):
        """
        Some files should never be cached.
        """
        response = make_response(send_from_directory("static/js", filename))
        response.headers[
            "Cache-Control"
        ] = "no-store, no-cache, must-revalidate, max-age=0"
        return response

    @app.errorhandler(500)
    def _internal_server_error(e):  # pylint: disable=unused-argument
        """
        Custom error handler for 500 Internal Server Error
        """
        exception_info = traceback.format_exc()
        # Should add logging ...
        # app.logger.error(exception_info)
        return (
            render_template(
                "errors/500_error.html",
                exception_info=exception_info,
                version=lute.__version__,
                platform=platform.platform(),
                is_docker=current_app.env_config.is_docker,
            ),
            500,
        )

    @app.errorhandler(404)
    def _page_not_found(e):  # pylint: disable=unused-argument
        "Show custom error page on 404."
        return (
            render_template(
                "errors/404_error.html",
                version=lute.__version__,
                requested_url=request.url,
                referring_page=request.referrer,
            ),
            404,
        )


def _create_app(app_config, extra_config):
    """
    Create the app using the given configuration,
    and init the SqlAlchemy db.
    """

    app = Flask(__name__, instance_path=app_config.datapath)

    config = {
        "SECRET_KEY": "some_secret",
        "DATABASE": app_config.dbfilename,
        "ENV": app_config.env,
        "SQLALCHEMY_DATABASE_URI": f"sqlite:///{app_config.dbfilename}",
        "DATAPATH": app_config.datapath,
        # ref https://flask-sqlalchemy.palletsprojects.com/en/2.x/config/
        # Don't track mods.
        "SQLALCHEMY_TRACK_MODIFICATIONS": False,
        # Disable CSRF -- this is a local app, and it's highly
        # unlikely that a malicious site will try to hack anyone's Lute data.
        # ref https://stackoverflow.com/questions/5207160/
        #   what-is-a-csrf-token-what-is-its-importance-and-how-does-it-work
        "WTF_CSRF_ENABLED": False,
    }

    final_config = {**config, **extra_config}
    app.config.from_mapping(final_config)

    # Attach the app_config to app so it's available at runtime.
    app.env_config = app_config

    db.init_app(app)

    @listens_for(Pool, "connect")
    def _pragmas_on_connect(dbapi_con, con_record):  # pylint: disable=unused-argument
        dbapi_con.execute("pragma recursive_triggers = on;")
        dbapi_con.execute("pragma foreign_keys = on;")

    with app.app_context():
        db.create_all()
        add_default_user_settings(db.session, app_config.default_user_backup_path)
        refresh_global_settings(db.session)
    app.db = db

    _add_base_routes(app, app_config)
    app.register_blueprint(language_bp)
    app.register_blueprint(book_bp)
    app.register_blueprint(bookmarks_bp)
    app.register_blueprint(term_bp)
    app.register_blueprint(termtag_bp)
    app.register_blueprint(read_bp)
    app.register_blueprint(bing_bp)
    app.register_blueprint(userimage_bp)
    app.register_blueprint(useraudio_bp)
    app.register_blueprint(termimport_bp)
    app.register_blueprint(backup_bp)
    app.register_blueprint(settings_bp)
    app.register_blueprint(themes_bp)
    app.register_blueprint(stats_bp)
    app.register_blueprint(cli_bp)
    if app_config.is_test_db:
        app.register_blueprint(dev_api_bp)

    return app


def _init_parser_plugins(plugin_data_path, outfunc):
    "Load and init plugins."
    outfunc("Initializing parsers from plugins ...")
    init_parser_plugins()

    parsers = supported_parsers()
    parsers_with_extra_data = [
        (typename, klass) for typename, klass in parsers if klass.uses_data_directory()
    ]
    if len(parsers_with_extra_data) > 0:
        # outfunc("Creating data folders for plugins ...")
        _setup_app_dir(plugin_data_path, "Data files for plugins.")
    for pair in parsers_with_extra_data:
        typename, klass = pair
        dirname = os.path.join(plugin_data_path, typename)
        klass.data_directory = dirname

        readme_content = f"Extra data for {klass.name()} plugin."
        _setup_app_dir(dirname, readme_content)
        klass.init_data_directory()
        # outfunc(f"  * {klass.name()}: {dirname}")

    outfunc("Enabled parsers:")
    for _, v in supported_parsers():
        outfunc(f"  * {v.name()}")


def create_app(
    app_config_path=None,
    extra_config=None,
    output_func=None,
):
    """
    App factory.  Calls dbsetup, and returns Flask app.

    Args:
    - app_config_path: path to yml file.  If None, use root config or default.
    - extra_config: dict, e.g. pass { 'TESTING': True } during unit tests.
    """

    def null_print(s):  # pylint: disable=unused-argument
        pass

    outfunc = output_func or null_print

    if app_config_path is None:
        if os.path.exists("config.yml"):
            app_config_path = "config.yml"
        else:
            app_config_path = AppConfig.default_config_filename()

    app_config = AppConfig(app_config_path)
    _setup_app_dirs(app_config)
    setup_db(app_config, output_func)

    if extra_config is None:
        extra_config = {}
    outfunc("Initializing app.")
    app = _create_app(app_config, extra_config)

    # Plugins are loaded after the app, as they may use settings etc.
    _init_parser_plugins(app_config.plugin_datapath, outfunc)

    return app


def data_initialization(session, output_func=None):
    """
    Any extra data setup.

    TODO: rework data initialization.  The DB setup can be handled
    outside of the application context, as IMO it's clearer to manage
    the data separately from the thing that uses the data.  This
    requires moving from flask-sqlalchemy to plain sqlalchemy.
    """

    def _null_print(s):  # pylint: disable=unused-argument
        pass

    outfunc = output_func or _null_print

    demosvc = DemoService(session)
    if demosvc.should_load_demo_data():
        outfunc("Loading demo data.")
        demosvc.load_demo_data()

    # TODO valid parsers: do parser check, mark valid as active, invalid as inactive.

    clean_data(session, outfunc)



"""
TextBookmark DataTable tests.
"""

import pytest
from lute.bookmarks.datatables import get_data_tables_list
from lute.db import db


@pytest.fixture(name="_dt_params")
def fixture_dt_params():
    "Sample query params."
    columns = [
        {"data": "0", "name": "TxOrder", "searchable": True, "orderable": True},
        {"data": "1", "name": "TbTitle", "searchable": True, "orderable": True},
    ]
    params = {
        "draw": "1",
        "columns": columns,
        "order": [{"column": "1", "dir": "asc"}],
        "start": "1",
        "length": "10",
        "search": {"value": "", "regex": False},
    }
    return params


def test_smoke_term_datatables_query_runs(app_context, _dt_params):
    """
    Smoke test only, ensure query runs.
    """
    data = get_data_tables_list(_dt_params, 1, db.session)
    assert data is not None



"""
Term entity.
"""

from lute.db import db

wordparents = db.Table(
    "wordparents",
    db.Model.metadata,
    db.Column("WpWoID", db.Integer, db.ForeignKey("words.WoID")),
    db.Column("WpParentWoID", db.Integer, db.ForeignKey("words.WoID")),
)


class TermImage(db.Model):
    "Term images."
    __tablename__ = "wordimages"

    id = db.Column("WiID", db.Integer, primary_key=True)
    term_id = db.Column(
        "WiWoID", db.Integer, db.ForeignKey("words.WoID"), nullable=False
    )
    term = db.relationship("Term", back_populates="images")
    source = db.Column("WiSource", db.String(500))


class TermFlashMessage(db.Model):
    "Term flash messages."
    __tablename__ = "wordflashmessages"

    id = db.Column("WfID", db.Integer, primary_key=True)
    message = db.Column("WfMessage", db.String(200))
    term_id = db.Column(
        db.Integer, db.ForeignKey("words.WoID"), name="WfWoID", nullable=False
    )
    term = db.relationship("Term", back_populates="term_flash_message", uselist=False)


wordtags = db.Table(
    "wordtags",
    db.Model.metadata,
    db.Column("WtTgID", db.Integer, db.ForeignKey("tags.TgID")),
    db.Column("WtWoID", db.Integer, db.ForeignKey("words.WoID")),
)


class TermTag(db.Model):
    "Term tags."
    __tablename__ = "tags"

    id = db.Column("TgID", db.Integer, primary_key=True)
    text = db.Column("TgText", db.String(20))
    _comment = db.Column("TgComment", db.String(200))

    def __init__(self, text, comment=None):
        self.text = text
        self.comment = comment

    @property
    def comment(self):
        "Comment getter."
        return self._comment

    # TODO zzfuture fix: TgComment should be nullable
    # The current schema has the TgComment NOT NULL default '',
    # which was a legacy copy over from LWT. Really, this should
    # be a nullable column, and all blanks should be NULL.
    # Minor change, a nice-to-have.
    @comment.setter
    def comment(self, c):
        "Set cleaned comment."
        self._comment = c if c is not None else ""


class TermTextChangedException(Exception):
    """
    Terms cannot change their text once saved,
    except for the text case.
    """


class Term(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Term entity.
    """

    __tablename__ = "words"

    id = db.Column("WoID", db.SmallInteger, primary_key=True)
    language_id = db.Column("WoLgID", db.Integer, db.ForeignKey("languages.LgID"))

    # Text should only be set through setters.
    _text = db.Column("WoText", db.String(250))

    # text_lc shouldn't be touched (it's changed when term.text is
    # set), but it's public here to allow for its access during
    # queries (eg in lute.read.service.  This can probably be
    # improved, not sure how at the moment.
    text_lc = db.Column("WoTextLC", db.String(250))

    status = db.Column("WoStatus", db.Integer)
    translation = db.Column("WoTranslation", db.String(500))
    romanization = db.Column("WoRomanization", db.String(100))
    token_count = db.Column("WoTokenCount", db.Integer)
    sync_status = db.Column("WoSyncStatus", db.Boolean)

    language = db.relationship("Language")
    term_tags = db.relationship("TermTag", secondary="wordtags")
    parents = db.relationship(
        "Term",
        secondary="wordparents",
        primaryjoin="Term.id == wordparents.c.WpWoID",
        secondaryjoin="Term.id == wordparents.c.WpParentWoID",
        back_populates="children",
    )
    children = db.relationship(
        "Term",
        secondary="wordparents",
        primaryjoin="Term.id == wordparents.c.WpParentWoID",
        secondaryjoin="Term.id == wordparents.c.WpWoID",
        back_populates="parents",
    )
    images = db.relationship(
        "TermImage",
        back_populates="term",
        lazy="subquery",
        cascade="all, delete-orphan",
    )
    term_flash_message = db.relationship(
        "TermFlashMessage",
        uselist=False,
        back_populates="term",
        cascade="all, delete-orphan",
    )

    def __init__(self, language=None, text=None):
        self.status = 1
        self.translation = None
        self.romanization = None
        self.sync_status = False
        self.term_tags = []
        self.parents = []
        self.children = []
        self.images = []
        if language is not None:
            self.language = language
        if text is not None:
            self.text = text

    @staticmethod
    def create_term_no_parsing(language, text):
        """
        Create a term, but do not reparse it during creation.

        This method is necessary because some parsers return
        different parsed tokens for a given text string based
        on its context.  The general __init__() is used for
        parsing without context, such as creating Terms from
        the UI or during CSV import.  This method is used
        when new terms are created from an already-parsed
        and already-tokenized page of text.
        """
        t = Term()
        t.language = language
        t._text = text  # pylint: disable=protected-access
        t.text_lc = language.get_lowercase(text)
        t.romanization = language.parser.get_reading(text)
        t._calc_token_count()  # pylint: disable=protected-access
        return t

    def __repr__(self):
        return f"<Term {self.id} '{self.text}'>"

    def __eq__(self, other):
        def eqrep(term):
            return f"{term.language.name} {term.text}"

        if isinstance(other, Term):
            return eqrep(self) == eqrep(other)
        return False

    @property
    def text(self):
        "Get the text."
        return self._text

    def _parse_string_add_zws(self, lang, textstring):
        "Parse the string using the language."
        # Clean up encoding cruft.
        t = textstring.strip()
        zws = "\u200B"  # zero-width space
        t = t.replace(zws, "")
        nbsp = "\u00A0"  # non-breaking space
        t = t.replace(nbsp, " ")

        tokens = lang.get_parsed_tokens(t)

        # Terms can't contain paragraph markers.
        tokens = [tok for tok in tokens if tok.token != "¶"]
        tok_strings = [tok.token for tok in tokens]

        t = zws.join(tok_strings)
        return t

    @text.setter
    def text(self, textstring):
        """
        Set the text, textlc, and token count.

        For new terms, just parse, downcase, and get the count.

        For existing terms, ensure that the actual text content has
        not changed.
        """
        if self.language is None:
            raise RuntimeError("Must set term language before setting text")
        lang = self.language

        if self.id is None:
            t = self._parse_string_add_zws(lang, textstring)
            self._text = t
            self.text_lc = lang.get_lowercase(t)
            self.romanization = lang.parser.get_reading(t)
            self._calc_token_count()
        else:
            # new_lc = lang.get_lowercase(textstring)
            # print(f"new lowercase = '{new_lc}', old = '{self.text_lc}'", flush=True)
            if lang.get_lowercase(textstring) != self.text_lc:
                msg = f'Cannot change text of saved term "{self._text}" (id {self.id}).'
                raise TermTextChangedException(msg)
            self._text = textstring

    def _calc_token_count(self):
        "Tokens are separated by zero-width space."
        token_count = 0
        if self._text is not None:
            zws = "\u200B"  # zero-width space
            parts = self._text.split(zws)
            token_count = len(parts)
        self.token_count = token_count

    def remove_all_term_tags(self):
        self.term_tags = []

    def add_term_tag(self, term_tag):
        if term_tag not in self.term_tags:
            self.term_tags.append(term_tag)

    def remove_term_tag(self, term_tag):
        if term_tag in self.term_tags:
            self.term_tags.remove(term_tag)

    def remove_all_parents(self):
        self.parents = []

    def add_parent(self, parent):
        """
        Add valid parent, term cannot be its own parent.
        """
        if self == parent:
            return
        if parent not in self.parents:
            self.parents.append(parent)
        if len(self.parents) > 1:
            self.sync_status = False

    def get_current_image(self, strip_jpeg=True):
        "Get the current (first) image for the term."
        if len(self.images) == 0:
            return None
        i = self.images[0]

        src = i.source

        if not strip_jpeg:
            return src

        # Ugly hack: we have to remove the .jpeg at the end because
        # Flask doesn't handle params with periods.
        return src.replace(".jpeg", "")

    def set_current_image(self, s):
        "Set the current image for this term."
        while len(self.images) > 0:
            self.images.pop(0)
        if (s or "").strip() != "":
            ti = TermImage()
            ti.term = self
            ti.source = s.strip()
            self.images.append(ti)

    def get_flash_message(self):
        "Get the flash message."
        if not self.term_flash_message:
            return None
        return self.term_flash_message.message

    def set_flash_message(self, m):
        "Set a flash message to be shown at some point in the future."
        tfm = self.term_flash_message
        if not tfm:
            tfm = TermFlashMessage()
            self.term_flash_message = tfm
        tfm.message = m

    def pop_flash_message(self):
        "Get the flash message, and remove it from this term."
        if not self.term_flash_message:
            return None
        m = self.term_flash_message.message
        self.term_flash_message = None
        return m


class Status(db.Model):  # pylint: disable=too-few-public-methods
    """
    Term statuses.
    """

    UNKNOWN = 0
    WELLKNOWN = 99
    IGNORED = 98

    __tablename__ = "statuses"

    id = db.Column("StID", db.SmallInteger, primary_key=True)
    text = db.Column("StText", db.String(250))
    abbreviation = db.Column("StAbbreviation", db.String(250))



"""
/termtag routes.
"""

from sqlalchemy import text
from flask import Blueprint, request, jsonify, render_template, redirect
from lute.models.term import TermTag
from lute.models.repositories import TermTagRepository
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.termtag.datatables import get_data_tables_list
from lute.db import db
from lute.termtag.forms import TermTagForm

bp = Blueprint("termtag", __name__, url_prefix="/termtag")


@bp.route("/index", defaults={"search": None}, methods=["GET"])
@bp.route("/index/<search>", methods=["GET"])
def index(search):
    "Index page."
    return render_template("termtag/index.html", initial_search=search)


@bp.route("/datatables", methods=["POST"])
def datatables_active_source():
    "Datatables data for terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    data = get_data_tables_list(parameters, db.session)
    return jsonify(data)


def _handle_form(termtag, form_template_name):
    """
    Handle a form post.
    """
    form = TermTagForm(obj=termtag)

    if form.validate_on_submit():
        form.populate_obj(termtag)
        db.session.add(termtag)
        db.session.commit()
        return redirect("/termtag/index", 302)

    return render_template(form_template_name, form=form, termtag=termtag)


@bp.route("/edit/<int:termtagid>", methods=["GET", "POST"])
def edit(termtagid):
    """
    Edit a termtag
    """
    repo = TermTagRepository(db.session)
    termtag = repo.find(termtagid)
    return _handle_form(termtag, "termtag/edit.html")


@bp.route("/new", methods=["GET", "POST"])
def new():
    """
    Create a termtag.
    """
    tt = TermTag("")
    return _handle_form(tt, "termtag/new.html")


@bp.route("/delete/<int:termtagid>", methods=["POST"])
def delete(termtagid):
    """
    Delete a termtag.
    """
    repo = TermTagRepository(db.session)
    termtag = repo.find(termtagid)
    db.session.delete(termtag)

    # ANNOYING HACK.  Per GitHub issue 455, the records
    # in the wordtags table were not getting deleted when
    # the tags were deleted with this method in the UI,
    # even though they were getting deleted during unit
    # testing.  I spent a short time looking, but am adding
    # this hack for now.
    # TODO fix_relationships: have wordtags records get deleted.
    sql = f"""delete from wordtags where WtTgID = {termtagid}"""
    db.session.execute(text(sql))

    db.session.commit()
    return redirect("/termtag/index", 302)



"""
Lute settings, in settings key-value table.
"""

import datetime
import time
from lute.db import db


class SettingBase(db.Model):
    """
    Main settings table.

    This class should not be used, it's the polymorphic base
    for UserSettings and SystemSettings.
    """

    __tablename__ = "settings"
    key = db.Column("StKey", db.String(40), primary_key=True)
    keytype = db.Column("StKeyType", db.String(10), nullable=False)
    value = db.Column("StValue", db.String, nullable=False)
    __mapper_args__ = {"polymorphic_on": keytype}


class UserSetting(SettingBase):
    "User setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "user"}


class BackupSettings:
    """
    Convenience wrapper for current backup settings.
    Getter only.
    """

    def __init__(self):
        self.backup_enabled = None
        self.backup_auto = None
        self.backup_warn = None
        self.backup_dir = None
        self.backup_count = None
        self.last_backup_datetime = None

    @property
    def last_backup_display_date(self):
        "Return the last_backup_datetime as yyyy-mm etc., or None if not set."
        t = self.last_backup_datetime
        if t is None:
            return None
        return datetime.datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S")

    @property
    def time_since_last_backup(self):
        """
        Return the time since the last backup. Returns None either if not set or
        it is in the future.
        Eg. "3 days ago"
        """
        t = self.last_backup_datetime
        if t is None:
            return None

        delta = int(time.time() - t)
        if delta < 0:
            return None

        thresholds = [
            ("week", 1 * 60 * 60 * 24 * 7),
            ("day", 1 * 60 * 60 * 24),
            ("hour", 1 * 60 * 60),
            ("minute", 1 * 60),
            ("second", 1),
        ]

        for unit, seconds in thresholds:
            multiples = abs(delta // seconds)
            if multiples >= 1:
                message = f"{multiples} {unit}"
                if multiples > 1:
                    message += "s"
                break
        else:
            message = f"{abs(delta)} seconds"

        return message + " ago"


class SystemSetting(SettingBase):
    "System setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "system"}



"""
Parsing using pythainlp

Includes classes:

- ThaiParser

"""

import re
import os
import pythainlp

from typing import List

from lute.parse.base import ParsedToken, AbstractParser


class ThaiParser(AbstractParser):
    """
    A parser for Thai that uses the pythainlp library for text segmentation.

    The user can add some exceptions to the "parsing_exceptions.txt"
    data file.
    """

    @classmethod
    def name(cls):
        return "Lute Thai"

    @classmethod
    def uses_data_directory(cls):
        "Uses the data_directory (defined in the AbstractParser)."
        return False

    # @classmethod
    # def init_data_directory(cls):
    #     "Set up necessary files."
    #     pass

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """
        text = text.replace("\r\n", "\n")

        words = pythainlp.word_tokenize(text)
        tokens = []
        pattern = f"[{language.word_characters}]"
        whitespace_regex = r"[ \t]+"
        for word in words:
            is_word_char = re.match(pattern, word) is not None
            is_whitespace = re.match(whitespace_regex, word) is not None
            is_split_sentence = word in language.regexp_split_sentences
            is_end_of_sentence = is_split_sentence or is_whitespace
            if is_end_of_sentence:
                is_word_char = False
            if word == "\n":
                word = "¶"
            if word == "¶":
                is_word_char = False
                is_end_of_sentence = True
            t = ParsedToken(word, is_word_char, is_end_of_sentence)
            tokens.append(t)
        return tokens

    def get_reading(self, text: str):  # pylint: disable=unused-argument
        """
        Get the pronunciation for the given text.  For most
        languages, this can't be automated.
        """
        return None



"""
Calculating stats.
"""

from datetime import datetime, timedelta
from sqlalchemy import text


def _get_data_per_lang(session):
    "Return dict of lang name to dict[date_yyyymmdd}: count"
    ret = {}
    sql = """
    select lang, dt, sum(WrWordCount) as count
    from (
      select LgName as lang, strftime('%Y-%m-%d', WrReadDate) as dt, WrWordCount
      from wordsread
      inner join languages on LgID = WrLgID
    ) raw
    group by lang, dt
    """
    result = session.execute(text(sql)).all()
    for row in result:
        langname = row[0]
        if langname not in ret:
            ret[langname] = {}
        ret[langname][row[1]] = int(row[2])
    return ret


def _charting_data(readbydate):
    "Calc data and running total."
    dates = sorted(readbydate.keys())
    if len(dates) == 0:
        return []

    # The line graph needs somewhere to start from for a line
    # to be drawn on the first day.
    first_date = datetime.strptime(dates[0], "%Y-%m-%d")
    day_before_first = first_date - timedelta(days=1)
    dbf = day_before_first.strftime("%Y-%m-%d")
    data = [{"readdate": dbf, "wordcount": 0, "runningTotal": 0}]

    total = 0
    for d in dates:
        dcount = readbydate.get(d)
        total += dcount
        hsh = {"readdate": d, "wordcount": dcount, "runningTotal": total}
        data.append(hsh)
    return data


def get_chart_data(session):
    "Get data for chart for each language."
    raw_data = _get_data_per_lang(session)
    chartdata = {}
    for k, v in raw_data.items():
        chartdata[k] = _charting_data(v)
    return chartdata


def _readcount_by_date(readbydate):
    """
    Return data as array: [ today, week, month, year, all time ]

    This may be inefficient, but will do for now.
    """
    today = datetime.now().date()

    def _in_range(i):
        start_date = today - timedelta(days=i)
        dates = [
            start_date + timedelta(days=x) for x in range((today - start_date).days + 1)
        ]
        ret = 0
        for d in dates:
            df = d.strftime("%Y-%m-%d")
            ret += readbydate.get(df, 0)
        return ret

    return {
        "day": _in_range(0),
        "week": _in_range(6),
        "month": _in_range(29),
        "year": _in_range(364),
        "total": _in_range(3650),  # 10 year drop off :-P
    }


def get_table_data(session):
    "Wordcounts by lang in time intervals."
    raw_data = _get_data_per_lang(session)

    ret = []
    for langname, readbydate in raw_data.items():
        ret.append({"name": langname, "counts": _readcount_by_date(readbydate)})
    return ret



"Stats service test."

from datetime import datetime, timedelta
from lute.models.book import WordsRead
from lute.db import db
from lute.stats.service import get_chart_data, get_table_data
from tests.utils import make_text


def make_read_text(lang, content, readdate):
    "Make and save a text."
    t = make_text(content, content, lang)
    # t.read_date = readdate
    db.session.add(t)
    db.session.commit()

    if readdate is None:
        return

    wr = WordsRead(t, readdate, t.word_count)
    db.session.add(wr)
    db.session.commit()


def test_get_chart_data(spanish, english, app_context):
    "Smoke test."
    today = datetime.now()
    yesterday = today - timedelta(days=1)
    daybefore = today - timedelta(days=2)

    make_read_text(spanish, "Yo tengo un gato.", today)
    make_read_text(spanish, "Ella esta aqui.", yesterday)
    make_read_text(spanish, "Nuevo text no leido.", None)
    make_read_text(english, "Yo yo.", today)

    expected = {
        "Spanish": [
            {
                "readdate": daybefore.strftime("%Y-%m-%d"),
                "wordcount": 0,
                "runningTotal": 0,
            },
            {
                "readdate": yesterday.strftime("%Y-%m-%d"),
                "wordcount": 3,
                "runningTotal": 3,
            },
            {"readdate": today.strftime("%Y-%m-%d"), "wordcount": 4, "runningTotal": 7},
        ],
        "English": [
            {
                "readdate": yesterday.strftime("%Y-%m-%d"),
                "wordcount": 0,
                "runningTotal": 0,
            },
            {"readdate": today.strftime("%Y-%m-%d"), "wordcount": 2, "runningTotal": 2},
        ],
    }
    assert get_chart_data(db.session) == expected


def test_get_table_data(spanish, english, app_context):
    "Smoke test."
    today = datetime.now()
    yesterday = today - timedelta(days=1)

    make_read_text(spanish, "Yo tengo un gato.", today)
    make_read_text(spanish, "Ella esta aqui.", yesterday)
    make_read_text(spanish, "Nuevo text no leido.", None)
    make_read_text(english, "Yo yo.", today)

    expected = [
        {
            "name": "English",
            "counts": {"day": 2, "week": 2, "month": 2, "year": 2, "total": 2},
        },
        {
            "name": "Spanish",
            "counts": {"day": 4, "week": 7, "month": 7, "year": 7, "total": 7},
        },
    ]
    actual = get_table_data(db.session)
    assert actual == expected


def test_get_data_works_when_nothing_read(app_context):
    "Nothing read should still be ok, empty chart."
    assert not get_chart_data(db.session), "nothing present"
    assert not get_table_data(db.session), "nothing"



"""
Term parent mapping.
"""

from lute.models.term import Term


## Exports


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    # TODO issue_336_export_unknown_book_terms: move this where needed.
    def export_unknown_terms(self, book, outfile):
        "Export unknown terms in the book to outfile."
        lang = book.language
        unique_tokens = {
            t
            for txt in book.texts
            for t in lang.get_parsed_tokens(txt.text)
            if t.is_word
        }
        unique_lcase_toks = {lang.get_lowercase(t.token) for t in unique_tokens}

        lgid = lang.id
        known_terms_lc = (
            self.session.query(Term.text_lc)
            .filter(Term.language_id == lgid, Term.token_count == 1)
            .all()
        )
        known_terms_lc = [word[0] for word in known_terms_lc]

        newtoks = [t for t in unique_lcase_toks if t not in known_terms_lc]
        with open(outfile, "w", encoding="utf-8") as f:
            f.write("\n".join(newtoks))



"""
Smoke test for bulk import of books.
"""

from sqlalchemy import and_
from lute.cli.import_books import import_books_from_csv
from lute.models.book import Book
from lute.models.repositories import BookRepository
from lute.db import db


def test_smoke_test(app_context, tmp_path, english, german):
    """Test importing books from CSV file"""
    csv_contents = """title,language,url,tags,audio,bookmarks,an extra column,text
A Book,English,http://www.example.com/book,"foo,bar,baz",book.mp3,1.00;3.14;42.00,extra information,"Lorem ipsum, dolor sit amet."
Another Book,,,,,,,The quick brown fox jumps over the lazy dog.
A Book,German,,,,,,Zwölf Boxkämpfer jagen Viktor quer über den großen Sylter Deich.
"""
    csv_file = tmp_path / "books.csv"
    with open(csv_file, "w", encoding="utf-8") as f:
        f.write(csv_contents)
    mp3_file = tmp_path / "book.mp3"
    with open(mp3_file, "w", encoding="utf-8") as f:
        pass

    common_tags = ["bar", "qux"]

    repo = BookRepository(db.session)

    # Check that no changes are made if not committing.
    import_books_from_csv(csv_file, "English", common_tags, False)
    assert repo.find_by_title("A Book", english.id) is None
    assert repo.find_by_title("Another Book", english.id) is None
    assert repo.find_by_title("A Book", german.id) is None

    # Check that new books are added.
    import_books_from_csv(csv_file, "English", common_tags, True)

    book = repo.find_by_title("A Book", english.id)
    assert book is not None
    assert book.title == "A Book"
    assert book.language_id == english.id
    assert book.source_uri == "http://www.example.com/book"
    assert book.audio_filename == str(mp3_file)
    assert book.audio_bookmarks == "1.00;3.14;42.00"
    assert len(book.texts) == 1
    assert book.texts[0].text == "Lorem ipsum, dolor sit amet."
    assert sorted([tag.text for tag in book.book_tags]) == ["bar", "baz", "foo", "qux"]

    book = repo.find_by_title("Another Book", english.id)
    assert book is not None
    assert book.title == "Another Book"
    assert book.language_id == english.id
    assert book.source_uri is None
    assert book.audio_filename is None
    assert book.audio_bookmarks is None
    assert len(book.texts) == 1
    assert book.texts[0].text == "The quick brown fox jumps over the lazy dog."
    assert sorted([tag.text for tag in book.book_tags]) == ["bar", "qux"]

    book = repo.find_by_title("A Book", german.id)
    assert book is not None
    assert book.title == "A Book"
    assert book.language_id == german.id
    assert book.source_uri is None
    assert book.audio_filename is None
    assert book.audio_bookmarks is None
    assert len(book.texts) == 1
    assert (
        book.texts[0].text
        == "Zwölf Boxkämpfer jagen Viktor quer über den großen Sylter Deich."
    )
    assert sorted([tag.text for tag in book.book_tags]) == ["bar", "qux"]

    # Check that duplicate books are not added.
    import_books_from_csv(csv_file, "English", common_tags, True)

    assert (
        db.session.query(Book)
        .filter(and_(Book.title == "A Book", Book.language_id == english.id))
        .count()
        == 1
    )
    assert (
        db.session.query(Book)
        .filter(and_(Book.title == "Another Book", Book.language_id == english.id))
        .count()
        == 1
    )
    assert (
        db.session.query(Book)
        .filter(and_(Book.title == "A Book", Book.language_id == german.id))
        .count()
        == 1
    )



"""
Term business object and repository.

Terms are converted to and from lute.models.term.Term objects to save
them in the database.
"""

import re
import sqlalchemy

from lute.models.term import Term as DBTerm, TermTag
from lute.models.repositories import (
    LanguageRepository,
    TermRepository,
    TermTagRepository,
)


class Term:  # pylint: disable=too-many-instance-attributes
    """
    Term business object.  All class members are primitives.
    """

    def __init__(self):
        # The ID of the DBTerm.
        self.id = None
        self.language_id = None
        # The text.
        self.text = None
        self.text_lc = None
        # The original text given to the DTO, to track changes.
        self.original_text = None
        self.status = 1
        self.translation = None
        self.romanization = None
        self.sync_status = False
        self.term_tags = []
        self.flash_message = None
        self.parents = []
        self.current_image = None

    def __repr__(self):
        return (
            f'<Term BO "{self.text}" lang_id={self.language_id} lang={self.language}>'
        )


class TermReference:
    "Where a Term has been used in books."

    def __init__(
        self, bookid, txid, pgnum, title, sentence=None
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self.book_id = bookid
        self.text_id = txid
        self.page_number = pgnum
        self.title = title
        self.sentence = sentence


class Repository:
    """
    Maps Term BO to and from lute.model.Term.
    """

    def __init__(self, _session):
        self.session = _session

        # Identity map for business lookup.
        # Note that the same term is stored
        # a few times: by upper case or lower case text,
        # plus language ID.
        self.identity_map = {}

    def _id_map_key(self, langid, text):
        return f"key-{langid}-{text}"

    def _add_to_identity_map(self, t):
        "Add found term to identity map."
        for txt in t.text, t.text_lc:
            k = self._id_map_key(t.language_id, txt)
            self.identity_map[k] = t

    def _search_identity_map(self, langid, txt):
        "Search the map w/ the given bus. key."
        k = self._id_map_key(langid, txt)
        if k in self.identity_map:
            return self.identity_map[k]
        return None

    def load(self, term_id):
        "Loads a Term business object for the DBTerm with the id."
        dbt = self.session.get(DBTerm, term_id)
        if dbt is None:
            raise ValueError(f"No term with id {term_id} found")
        term = self._build_business_term(dbt)
        self._add_to_identity_map(term)
        return term

    def _search_spec_term(self, langid, text):
        """
        Make a term to get the correct text_lc to search for.
        This ensures that the spec term is properly parsed
        and downcased.
        """
        lang_repo = LanguageRepository(self.session)
        lang = lang_repo.find(langid)
        return DBTerm(lang, text)

    def _find_by_spec(self, langid, text):
        "Do a search using a spec term."
        spec = self._search_spec_term(langid, text)
        repo = TermRepository(self.session)
        return repo.find_by_spec(spec)

    def find(self, langid, text):
        """
        Return a Term business object for the DBTerm with the langid and text.
        If no match, return None.
        """
        term = self._search_identity_map(langid, text)
        if term is not None:
            return term

        dbt = self._find_by_spec(langid, text)
        if dbt is None:
            return None
        term = self._build_business_term(dbt)
        self._add_to_identity_map(term)
        return term

    def find_or_new(self, langid, text):
        """
        Return a Term business object for the DBTerm with the langid and text.
        If no match, return a new term with the text and language.

        Note that this does a search by the **tokenized version**
        of the text; i.e., first the text argument is converted into
        a "search specification" (spec) using the language with the given id.
        The db search is then done using this spec.  In most cases, this
        # will suffice.

        # In some cases, though, it may cause errors.  The parsing here is done
        # without a fuller context, which in some language parsers can result
        # in different results.  For example, the Japanese "集めれ" string can
        # can be parsed with mecab to return one unit ("集めれ") or two ("集め/れ"),
        # depending on context.

        # So what does this mean?  It means that any context-less searches
        # for terms that have ambiguous parsing results will, themselves,
        # also be ambiguous.  This impacts csv imports and term form usage.

        # For regular (reading screen) usage, it probably doesn't matter.
        # The terms in the reading screen are all created when the page is
        # opened, and so have ids assigned.  With that, terms are not
        # searched by text match, they are only searched by id.

        ## TODO verify_identity_map_comment:
        If it's new, don't add to the identity map ... it's not saved yet,
        and so if we search for it again we should hit the db again.

        # the above statement about the identity map was old code, and I'm not
        # sure it's a valid statement/condition.
        """
        t = self.find(langid, text)
        if t is not None:
            return t

        spec = self._search_spec_term(langid, text)
        t = Term()
        t.language_id = langid
        t.text = spec.text
        t.text_lc = spec.text_lc
        t.romanization = spec.language.parser.get_reading(text)
        t.original_text = spec.text

        # TODO verify_identity_map_comment
        # Adding the term to the map, even though it's new.
        self._add_to_identity_map(t)

        return t

    def find_matches(self, langid, text, max_results=50):
        """
        Return array of Term business objects for the DBTerms
        with the same langid, matching the text.
        If no match, return [].
        """
        spec = self._search_spec_term(langid, text)
        text_lc = spec.text_lc
        search = text_lc.strip() if text_lc else ""
        if search == "":
            return []

        sql_query = """SELECT
        t.WoID as id,
        t.WoText as text,
        t.WoTextLC as text_lc,
        t.WoTranslation as translation,
        t.WoStatus as status,
        t.WoLgID as language_id,
        CASE WHEN wp.WpParentWoID IS NOT NULL THEN 1 ELSE 0 END AS has_children,
        CASE WHEN t.WoTextLC = :text_lc THEN 2
          WHEN t.WoTextLC LIKE :text_lc_starts_with THEN 1
          ELSE 0
        END as text_starts_with_search_string

        FROM words AS t
        LEFT JOIN (
          select WpParentWoID from wordparents group by WpParentWoID
        ) wp on wp.WpParentWoID = t.WoID

        WHERE t.WoLgID = :langid AND t.WoTextLC LIKE :text_lc_wildcard

        ORDER BY text_starts_with_search_string DESC, has_children DESC, t.WoTextLC
        LIMIT :max_results
        """
        # print(sql_query)
        params = {
            "text_lc": text_lc,
            "text_lc_wildcard": f"%{text_lc}%",
            "text_lc_starts_with": f"{text_lc}%",
            "langid": langid,
            "max_results": max_results,
        }
        # print(params)

        alchsql = sqlalchemy.text(sql_query)
        return self.session.execute(alchsql, params).fetchall()

    def get_term_tags(self):
        "Get all available term tags, helper method."
        tags = self.session.query(TermTag).all()
        return sorted([t.text for t in tags])

    def add(self, term):
        """
        Add a term to be saved to the db session.
        Returns DB Term for tests and verification only,
        clients should not change it.
        """
        dbterm = self._build_db_term(term)
        self.session.add(dbterm)
        return dbterm

    def delete(self, term):
        """
        Add term to be deleted to session.
        """
        dbt = None
        if term.id is not None:
            dbt = self.session.get(DBTerm, term.id)
        else:
            dbt = self._find_by_spec(term.language_id, term.text)
        if dbt is not None:
            self.session.delete(dbt)

    def commit(self):
        """
        Commit everything, flush the map to force refetches.
        """
        self.identity_map = {}
        self.session.commit()

    def _build_db_term(self, term):
        "Convert a term business object to a DBTerm."
        # print(f"in _build_db_term, term id = {term.id}", flush=True)
        if term.text is None:
            raise ValueError("Text not set for term")

        t = None
        if term.id is not None:
            # This is an existing term, so use it directly.
            t = self.session.get(DBTerm, term.id)
        else:
            # New term, or finding by text.
            spec = self._search_spec_term(term.language_id, term.text)
            term_repo = TermRepository(self.session)
            t = term_repo.find_by_spec(spec) or DBTerm()
            t.language = spec.language

        t.text = term.text
        t.original_text = term.text
        t.status = term.status
        t.translation = term.translation
        t.romanization = term.romanization
        t.sync_status = term.sync_status
        t.set_current_image(term.current_image)

        if term.flash_message is not None:
            t.set_flash_message(term.flash_message)
        else:
            t.pop_flash_message()

        tt_repo = TermTagRepository(self.session)
        termtags = []
        for s in list(set(term.term_tags)):
            termtags.append(tt_repo.find_or_create_by_text(s))
        t.remove_all_term_tags()
        for tt in termtags:
            t.add_term_tag(tt)

        termparents = []
        lang = t.language
        create_parents = [
            p
            for p in term.parents
            if p is not None
            and p != ""
            and lang.get_lowercase(term.text) != lang.get_lowercase(p)
        ]
        # print('creating parents: ' + ', '.join(create_parents))
        for p in create_parents:
            termparents.append(self._find_or_create_parent(p, lang, term, termtags))
        t.remove_all_parents()
        for tp in termparents:
            t.add_parent(tp)

        if len(termparents) != 1:
            t.sync_status = False

        # print(f"in _build_db_term, returning db term with term id = {t.id}", flush=True)
        return t

    def _find_or_create_parent(self, pt, language, term, termtags) -> DBTerm:
        p = self._find_by_spec(language.id, pt)
        new_or_unknown_parent = p is None or p.status == 0
        new_term = term.id is None

        if p is None:
            p = DBTerm(language, pt)

        if new_or_unknown_parent:
            p.status = term.status

        # Copy translation, image if missing, but _not_ if we're just
        # re-saving an existing term.
        if new_or_unknown_parent or new_term:
            if (p.translation or "") == "":
                p.translation = term.translation
            if (p.get_current_image() or "") == "":
                p.set_current_image(term.current_image)

        # Only copy tags if this is a new parent.  New parents should
        # _likely_ inherity the tags of the term.
        if new_or_unknown_parent:
            for tt in termtags:
                p.add_term_tag(tt)

        return p

    def _build_business_term(self, dbterm):
        "Create a Term bus. object from a lute.model.term.Term."
        term = Term()
        term.id = dbterm.id
        term.language_id = dbterm.language.id

        text = dbterm.text
        ## Remove zero-width spaces (zws) from strings for user forms.
        #
        # NOTE: disabling this as it creates challenges for editing
        # terms.  In some cases, the same term may have a zws
        # character as part of it; in other cases, it won't, e.g. "
        # 集めれ" sometimes is parsed as one token, and sometimes
        # two ("集め/れ").  If we strip the zws from the string, then
        # when it's posted back, Lute will think that it has changed.
        # ... it gets messy.
        # zws = "\u200B"  # zero-width space
        # text = text.replace(zws, "")
        term.text_lc = dbterm.text_lc
        term.original_text = text
        term.text = text

        term.status = dbterm.status
        term.translation = dbterm.translation
        term.romanization = dbterm.romanization
        term.sync_status = dbterm.sync_status
        term.current_image = dbterm.get_current_image()
        term.flash_message = dbterm.get_flash_message()
        term.parents = [p.text for p in dbterm.parents]
        term.romanization = dbterm.romanization
        term.term_tags = [tt.text for tt in dbterm.term_tags]

        return term

    ## References.

    def find_references(self, term):
        """
        Return references of term, children, and parents.
        """
        spec = self._search_spec_term(term.language_id, term.text)
        term_repo = TermRepository(self.session)
        searchterm = term_repo.find_by_spec(spec)
        if searchterm is None:
            searchterm = spec

        references = {
            "term": self._get_references(searchterm),
            "children": self._get_child_references(searchterm),
            "parents": self._get_parent_references(searchterm),
        }
        return references

    def _build_term_references(self, term_lc, rows):
        ret = []
        zws = chr(0x200B)  # zero-width space
        for row in rows:
            sentence = row[4].strip()
            pattern = f"{zws}({term_lc}){zws}"

            def replace_match(m):
                return f"{zws}<b>{m.group(0)}</b>{zws}"

            sentence = re.sub(pattern, replace_match, sentence, flags=re.IGNORECASE)
            sentence = sentence.replace(zws, "").replace("¶", "")
            ret.append(TermReference(row[0], row[1], row[2], row[3], sentence))
        return ret

    def _get_references(self, term):
        """
        Search the sentences.text_content (or textlc_content if needed).

        sentence.textlc_content is set to '*' if a call to sqlite's LOWER
        returns the same data as using the sentence Language.parser.  This
        saves a pile of space, at least in my case with Spanish, as only
        0.5% of the lowercased sentences actually differ.
        """
        if term is None:
            return []

        term_lc = term.text_lc
        query = sqlalchemy.text(
            f"""
            SELECT DISTINCT
                texts.TxBkID,
                TxID,
                TxOrder,
                BkTitle || ' (' || TxOrder || '/' || pc.c || ')' AS TxTitle,
                SeText
            FROM sentences
            INNER JOIN texts ON TxID = SeTxID
            INNER JOIN books ON BkID = texts.TxBkID
            INNER JOIN (
                SELECT TxBkID, COUNT(*) AS c
                FROM texts
                GROUP BY TxBkID
            ) pc ON pc.TxBkID = texts.TxBkID
            WHERE TxReadDate IS NOT NULL
            AND SeText IS NOT NULL
            AND CASE WHEN SeTextLC == '*' THEN SeText ELSE SeTextLC END LIKE :pattern
            AND BkLgID = {term.language.id}
            LIMIT 20
        """
        )

        pattern = f"%{chr(0x200B)}{term_lc}{chr(0x200B)}%"
        params = {"pattern": pattern}
        result = self.session.execute(query, params)
        return self._build_term_references(term_lc, result)

    def _get_all_refs(self, terms):
        all_references = []
        for term in terms:
            references = self._get_references(term)
            all_references.extend(references)
        return all_references

    def _get_parent_references(self, term):
        parent_references = []
        for parent in term.parents:
            parent_term_lc = parent.text_lc
            references = self._get_family_references(parent, term)
            parent_references.append({"term": parent_term_lc, "refs": references})
        return parent_references

    def _get_family_references(self, parent, term):
        if term is None or parent is None:
            return []
        family = [parent] + parent.children
        family = [t for t in family if t.id != term.id]
        return self._get_all_refs(family)

    def _get_child_references(self, term):
        if term is None:
            return []
        return self._get_all_refs(term.children)



"""
Step defs for term_import.feature.
"""
# pylint: disable=missing-function-docstring

import os
import tempfile
import pytest

from pytest_bdd import given, when, then, scenarios, parsers

from lute.db import db
from lute.models.language import Language
from lute.language.service import Service as LanguageService
from lute.models.term import Term
from lute.models.repositories import LanguageRepository, TermRepository
from lute.termimport.service import Service, BadImportFileError

from tests.dbasserts import assert_sql_result


# The content of the file for the current test.
content = None

# The results of the import
stats = None

scenarios("term_import.feature")
scenarios("term_import_status_0.feature")


@given("demo data")
def given_demo_data(app_context):
    "Load languages necessary for imports."
    svc = LanguageService(db.session)
    for lang in ["Spanish", "English", "Classical Chinese"]:
        svc.load_language_def(lang)


@given(parsers.parse("import file:\n{newcontent}"))
def given_file(newcontent):
    global content  # pylint: disable=global-statement
    content = newcontent


@given("empty import file")
def given_empty_file():
    global content  # pylint: disable=global-statement
    content = ""


@when(parsers.parse("import with create {create}, update {update}"))
def import_with_settings(create, update):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)

    global stats  # pylint: disable=global-statement
    service = Service(db.session)
    stats = service.import_file(
        path, create.lower() == "true", update.lower() == "true"
    )
    os.remove(path)


@when(
    parsers.parse(
        "import with create {create}, update {update}, new as unknown {newunknowns}"
    )
)
def import_with_settings_and_newunks(create, update, newunknowns):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)

    global stats  # pylint: disable=global-statement
    service = Service(db.session)
    stats = service.import_file(
        path,
        create.lower() == "true",
        update.lower() == "true",
        newunknowns.lower() == "true",
    )
    os.remove(path)


@then(
    parsers.parse(
        "import should succeed with {created} created, {updated} updated, {skipped} skipped"
    )
)
def succeed_with_status(created, updated, skipped):
    assert stats["created"] == int(created), "created"
    assert stats["updated"] == int(updated), "updated"
    assert stats["skipped"] == int(skipped), "skipped"


@then(parsers.parse("import should fail with message:\n{message}"))
def fail_with_message(message):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)
    with pytest.raises(BadImportFileError, match=message):
        service = Service(db.session)
        service.import_file(path)
    os.remove(path)


@then(parsers.parse("words table should contain:\n{text_lc_content}"))
def then_words_table_contains_WoTextLC(text_lc_content):
    expected = []
    if text_lc_content != "-":
        expected = text_lc_content.split("\n")
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, expected)


@then(parsers.parse('{language} term "{term}" should be:\n{expected}'))
def then_term_tags(language, term, expected):
    repo = LanguageRepository(db.session)
    lang = repo.find_by_name(language)
    spec = Term(lang, term)
    term_repo = TermRepository(db.session)
    t = term_repo.find_by_spec(spec)
    pstring = ", ".join([p.text for p in t.parents])
    if pstring == "":
        pstring = "-"
    tstring = ", ".join(sorted([p.text for p in t.term_tags]))
    if tstring == "":
        tstring = "-"
    actual = [
        f"translation: {t.translation or '-'}",
        f"pronunciation: {t.romanization or '-'}",
        f"status: {t.status}",
        f"parents: {pstring}",
        f"tags: {tstring}",
    ]
    assert "\n".join(actual) == expected, "term"


@then(parsers.parse('sql "{sql}" should return:\n{expected}'))
def then_sql_returns(sql, expected):
    expected = expected.split("\n")
    assert_sql_result(sql, expected)



"""
TextBookmark mapping checks.
"""
from datetime import datetime
import pytest
from lute.models.book import Book, Text, TextBookmark
from lute.db import db
from tests.dbasserts import assert_record_count_equals, assert_sql_result


@pytest.fixture(name="sample_book")
def fixture_sample_book(english):
    "Sample Book"
    b = Book.create_book("Book Title", english, "some text")
    b.texts[0].read_date = datetime.now()
    return b


@pytest.fixture(name="sample_text")
def fixture_sample_text(sample_book: Book):
    "Sample Text"
    t = Text(sample_book, "test text", 1)
    return t


@pytest.fixture(name="sample_textbookmark")
def fixture_sample_bookmark(sample_text: Text):
    "Sample TextBookmark"
    tb = TextBookmark(text=sample_text, title="Test Title")
    return tb


def test_save_textbookmark(
    empty_db, sample_textbookmark: TextBookmark, sample_text: Text
):
    """Check TextBookmark mappings"""
    db.session.add(sample_text)
    db.session.add(sample_textbookmark)
    db.session.commit()
    sql = "SELECT TbTxId, TbTitle FROM textbookmarks"
    assert_sql_result(
        sql, [f"{sample_text.id}; {sample_textbookmark.title}"], "textbookmark"
    )


def test_edit_textbookmark(
    empty_db, sample_textbookmark: TextBookmark, sample_text: Text
):
    """Edit TextBookmark"""
    db.session.add(sample_text)
    db.session.add(sample_textbookmark)
    db.session.commit()

    db.session.query(TextBookmark).filter(
        TextBookmark.text.has(id=sample_text.id)
    ).update({"title": "New Title"})
    db.session.commit()

    sql = "SELECT TbTxId, TbTitle FROM textbookmarks"
    assert_sql_result(sql, [f"{sample_text.id}; New Title"], "textbookmark")


def test_delete_textbookmark(
    empty_db, sample_textbookmark: TextBookmark, sample_text: Text
):
    """Delete TextBookmark"""
    db.session.add(sample_text)
    db.session.add(sample_textbookmark)
    db.session.commit()

    db.session.query(TextBookmark).filter(
        TextBookmark.text.has(id=sample_text.id)
    ).delete()
    db.session.commit()

    assert_record_count_equals("textbookmarks", 0, "0 bookmarks")



"""
Parser registry.

List of available parsers.
"""

from importlib.metadata import entry_points
from sys import version_info

from lute.parse.base import AbstractParser
from lute.parse.space_delimited_parser import SpaceDelimitedParser, TurkishParser
from lute.parse.mecab_parser import JapaneseParser
from lute.parse.character_parser import ClassicalChineseParser


__LUTE_PARSERS__ = {
    "spacedel": SpaceDelimitedParser,
    "turkish": TurkishParser,
    "japanese": JapaneseParser,
    "classicalchinese": ClassicalChineseParser,
}


def init_parser_plugins():
    """
    Initialize parsers from plugins
    """

    # Handle API breakage of entry_points.
    # pylint: disable=no-member
    vmaj = version_info.major
    vmin = version_info.minor
    if vmaj == 3 and vmin in (8, 9, 10, 11):
        custom_parser_eps = entry_points().get("lute.plugin.parse")
    elif (vmaj == 3 and vmin >= 12) or (vmaj >= 4):
        # Can't be sure this will always work, API may change again,
        # but can't plan for the unforseeable everywhere.
        custom_parser_eps = entry_points().select(group="lute.plugin.parse")
    else:
        # earlier version of python than 3.8?  What madness is this?
        # Not going to throw, just print and hope the user sees it.
        msg = f"Unable to load plugins for python {vmaj}.{vmin}, please upgrade to 3.8+"
        print(msg, flush=True)
        return

    if custom_parser_eps is None:
        return

    for custom_parser_ep in custom_parser_eps:
        name = custom_parser_ep.name
        klass = custom_parser_ep.load()
        if issubclass(klass, AbstractParser):
            __LUTE_PARSERS__[name] = klass
        else:
            raise ValueError(f"{name} is not a subclass of AbstractParser")


def get_parser(parser_name) -> AbstractParser:
    "Return the supported parser with the given name."
    if parser_name not in __LUTE_PARSERS__:
        raise ValueError(f"Unknown parser type '{parser_name}'")
    pclass = __LUTE_PARSERS__[parser_name]
    if not pclass.is_supported():
        raise ValueError(f"Unsupported parser type '{parser_name}'")
    return pclass()


def is_supported(parser_name) -> bool:
    "Return True if the specified parser is present and supported."
    if parser_name not in __LUTE_PARSERS__:
        return False
    p = __LUTE_PARSERS__[parser_name]
    return p.is_supported()


def supported_parsers():
    """
    List of supported parser strings and classes.
    """
    return [(k, v) for k, v in __LUTE_PARSERS__.items() if v.is_supported()]


def supported_parser_types():
    """
    List of supported Language.parser_types
    """
    return list(a[0] for a in supported_parsers())



"""
Given text and Terms, determine what to render in the browser.

For example, given the following TextTokens A-I:

  A B C D E F G H I

And the following terms:

  "A" through "I" (single-word terms)
  "B C"       (term J)
  "E F G H I" (K)
  "F G"       (L)
  "C D E"     (M)

The following TextItems would be displayed on the reading screen,
with some of the Terms overlapping:

  [A][B C][-D E][-F G H I]
"""

import re
from collections import Counter
from lute.models.term import Term
from lute.read.render.text_item import TextItem

# from lute.utils.debug_helpers import DebugTimer

zws = "\u200B"  # zero-width space


def get_string_indexes(strings, content):
    """
    Returns list of arrays: [[string, index], ...]

    e.g., _get_string_indexes(["is a", "cat"], "here is a cat")
    returns [("is a", 1), ("cat", 3)].

    strings and content must be lowercased!
    """
    searchcontent = zws + content + zws
    zwsindexes = [index for index, letter in enumerate(searchcontent) if letter == zws]

    ret = []

    for s in strings:
        # "(?=())" is required because sometimes the search pattern can
        # overlap -- e.g. _b_b_ has _b_ *twice*.
        # https://stackoverflow.com/questions/5616822/
        #   how-to-use-regex-to-find-all-overlapping-matches
        pattern = rf"(?=({re.escape(zws + s + zws)}))"
        add_matches = [
            (s, zwsindexes.index(m.start()))
            for m in re.finditer(pattern, searchcontent)
        ]
        ret.extend(add_matches)

    return ret


# pylint: disable=too-many-arguments,too-many-positional-arguments
def _make_textitem(index, text, text_lc, count, sentence_number, term):
    "Make a TextItem."
    r = TextItem()
    r.text = text
    r.sentence_number = sentence_number
    r.text_lc = text_lc
    r.token_count = count
    r.display_count = r.token_count
    r.index = index
    r.is_word = term is not None
    r.term = term
    return r


def _create_missing_status_0_terms(tokens, terms, language):
    "Make new terms as needed for all tokens, using case of last instance."

    original_word_tokens = {t.token for t in tokens if t.is_word}
    parser = language.parser
    lc_word_tokens = {parser.get_lowercase(t): t for t in original_word_tokens}
    term_text_lcs = {t.text_lc for t in terms}

    missing_word_tokens = [
        original for lc, original in lc_word_tokens.items() if lc not in term_text_lcs
    ]

    # Note: create the terms _without parsing_ because some parsers
    # break up characters when the words are given out of context.
    missing_word_tokens = list(set(missing_word_tokens))
    new_terms = [Term.create_term_no_parsing(language, t) for t in missing_word_tokens]
    for t in new_terms:
        t.status = 0

    return new_terms


def get_textitems(tokens, terms, language, multiword_term_indexer=None):
    """
    Return TextItems that will **actually be rendered**.

    Method to determine what should be rendered:

    - Create TextItems for all of the tokens, finding their
      starting index in the tokens.

    - "Write" the TextItems to an array in correctly sorted
      order, so that the correct TextItems take precendence
      in the final rendering.

    - Calculate any term overlaps.

    - Return the final list of TextItems that will actually
      be rendered.

    ---

    Applying the above algorithm to the example given in the class
    header:

    We have the following TextTokens A-I:

      A B C D E F G H I

    And given the following terms:
      "A" through "I" (single-word terms)
      "B C"       (term J)
      "E F G H I" (K)
      "F G"       (L)
      "C D E"     (M)

    Creating TextItems for all of the terms, finding their starting
    indices in the tokens:

      TextToken    index   length
      ----         -----   ------
      [A]          0       1
      [B]          1       1
      ...
      [I]          8       1
      [B C]        1       2
      [E F G H I]  4       5
      [F G]        5       2
      [C D E]      2       3

    Sorting by index, then decreasing token count:

      TextToken    index   length   ID (for later reference)
      ----         -----   ------   ------------------------
      [A]          0       1        t1
      [B C]        1       2        t2
      [B]          1       1        t3
      [C D E]      2       3        t4
      [C]          2       1        t5
      [D]          3       1        t6
      [E F G H I]  4       5        t7
      [E]          4       1        t8
      [F G]        5       2        t9
      [F]          5       1        t10
      [G]          6       1        t11
      [H]          7       1        t12
      [I]          8       1        t13

    Starting at the bottom of the above list and
    working upwards:

      - ID of [I] is written to index 8: [] [] [] [] [] [] [] [] [t13]
      - ID of [H] to index 7: [] [] [] [] [] [] [] [t12] [t13]
      - ...
      - [F G] to index 5 *and* 6: [] [] [] [] [] [t9] [t9] [t12] [t13]
      - [E] to index 4:  [] [] [] [] [t8] [t9] [t9] [t12] [t13]
      - [E F G H I] to indexes 4-8:   [] [] [] [] [t7] [t7] [t7] [t7] [t7]
      - ... etc

    Using the TextItem IDs, the resulting array would be:

      output array: [t1] [t2] [t2] [t4] [t4] [t7] [t7] [t7] [t7]
                    [A]  [B     C] [-D    E] [-F    G    H    I]

    The only TextItems that will be shown are therefore:
      t1, t2, t3, t7

    To calculate what text is actually displayed, the count
    of each ID is used.  e.g.:
      - ID t7 appears 4 times in the output array.  The last 4 tokens of
        [E F G H I] are [F G H I], which will be used as t7's display text.
      - ID t2 appears 2 times.  The last 2 tokens of [B C] are [B C],
        so that will be the display text. etc.
    """
    # pylint: disable=too-many-locals

    # dt = DebugTimer("get_textitems", display=False)

    new_unknown_terms = _create_missing_status_0_terms(tokens, terms, language)
    # dt.step("new_unknown_terms")

    all_terms = terms + new_unknown_terms
    text_to_term = {dt.text_lc: dt for dt in all_terms}

    tokens_orig = [t.token for t in tokens]
    tokens_lc = [language.parser.get_lowercase(t) for t in tokens_orig]

    textitems = []

    def _add_textitem(index, text_lc, count):
        "Add a TextItem for position index in tokens."
        text_orig = tokens_orig[index]
        if count > 1:
            text_orig = zws.join(tokens_orig[index : index + count])
        text_lc = zws.join(tokens_lc[index : index + count])
        sentence_number = tokens[index].sentence_number
        term = text_to_term.get(text_lc, None)
        ti = _make_textitem(index, text_orig, text_lc, count, sentence_number, term)
        textitems.append(ti)

    # Single-word terms.
    for index, _ in enumerate(tokens):
        _add_textitem(index, tokens_lc[index], 1)
    # dt.step("single word textitems")

    # Multiword terms.
    if multiword_term_indexer is not None:
        for r in multiword_term_indexer.search_all(tokens_lc):
            mwt = text_to_term[r[0]]
            count = mwt.token_count
            _add_textitem(r[1], r[0], count)
        # dt.step(f"get mw textitems w indexer")
    else:
        multiword_terms = [t.text_lc for t in all_terms if t.token_count > 1]
        for e in get_string_indexes(multiword_terms, zws.join(tokens_lc)):
            count = e[0].count(zws) + 1
            _add_textitem(e[1], e[0], count)
        # dt.step("mw textitems without indexer")

    # Sorting by index, then decreasing token count.
    textitems = sorted(textitems, key=lambda x: (x.index, -x.token_count))

    # "Write out" TextItems to the output array.
    output_textitem_ids = [None] * len(tokens)
    for ti in reversed(textitems):
        for c in range(ti.index, ti.index + ti.token_count):
            output_textitem_ids[c] = id(ti)

    # Calc display_counts; e.g. if a textitem's id shows up 3 times
    # in the output_textitem_ids, it should display 3 tokens.
    id_counts = dict(Counter(output_textitem_ids))
    for ti in textitems:
        ti.display_count = id_counts.get(id(ti), 0)
    # dt.step("display_count")

    textitems = [ti for ti in textitems if ti.display_count > 0]

    current_paragraph = 0
    for ti in textitems:
        ti.paragraph_number = current_paragraph
        if ti.text == "¶":
            current_paragraph += 1
    # dt.step("paragraphs")
    # dt.step("done")

    return textitems



"TextItem class."

zws = "\u200B"  # zero-width space


class TextItem:  # pylint: disable=too-many-instance-attributes
    """
    Unit to be rendered.

    Data structure for template read/textitem.html

    Some elements are lazy loaded, because they're only needed in
    certain situations.
    """

    def __init__(self, term=None):
        self.index: int
        self.lang_id: int
        self.text: str  # The original, un-overlapped text.
        self.text_lc: str
        self.is_word: int

        # Number of tokens originally in the Text item.
        self.token_count: int = 1

        # Number of tokens that should be displayed, starting from the
        # end of the string.
        self.display_count: int = 1

        self.sentence_number: int = 0
        self.paragraph_number: int = 0

        # Calls setter
        self.term = term

        self.extra_html_classes = []

        # TODO code
        # # The flash message can be None, so we need an extra flag
        # # to determine if it has been loaded or not.
        # self._flash_message_loaded: bool = False
        # self._flash_message: str = None

    def __repr__(self):
        return f'<TextItem "{self.text}" (wo_id={self.wo_id}, sent={self.sentence_number})>'

    @property
    def term(self):
        return self._term

    @property
    def wo_id(self):
        "The term id is the wo_id."
        if self._term is None:
            return None
        return self._term.id

    @term.setter
    def term(self, t):
        self.wo_status = None
        self._term = t
        if t is None:
            return
        self.lang_id = t.language.id
        self.wo_status = t.status

    # TODO - reactivate with non-lazy query results.
    # @property
    # def flash_message(self):
    #     """
    #     Return flash message if anything present.
    #     Lazy loaded as needed.
    #     """
    #     if self._flash_message_loaded:
    #         return self._flash_message
    #     if self.term is None:
    #         return None

    #     self._flash_message = self.term.get_flash_message()
    #     self._flash_message_loaded = True
    #     return self._flash_message

    @property
    def display_text(self):
        "Show last n tokens, if some of the textitem is covered."
        toks = self.text.split(zws)
        disp_toks = toks[-self.display_count :]
        return zws.join(disp_toks)

    @property
    def html_display_text(self):
        """
        Content to be rendered to browser.
        """
        return self.display_text.replace(zws, "")

    @property
    def span_id(self):
        """
        Each span gets a unique id.
        """
        return f"ID-{self.sentence_number}-{self.index}"

    @property
    def status_class(self):
        "Status class to apply."
        if self.wo_id is None:
            return "status0"
        return f"status{self.wo_status}"

    def add_html_class(self, c):
        "Add extra class to term."
        self.extra_html_classes.append(c)

    @property
    def html_class_string(self):
        """
        Create class string for TextItem.
        """
        if self.is_word == 0:
            return "textitem"

        classes = [
            "textitem",
            "click",
            "word",
            "word" + str(self.wo_id),
        ]

        if self.display_text != self.text:
            classes.append("overlapped")
        classes.extend(self.extra_html_classes)

        return " ".join(classes)



"""
Find terms in contest string using ahocorapy.
"""

from ahocorapy.keywordtree import KeywordTree


class MultiwordTermIndexer:
    """
    Find terms in strings using ahocorapy.
    """

    zws = "\u200B"  # zero-width space

    def __init__(self):
        self.kwtree = KeywordTree(case_insensitive=True)
        self.finalized = False

    def add(self, t):
        "Add zws-enclosed term to tree."
        add_t = f"{self.zws}{t}{self.zws}"
        self.kwtree.add(add_t)

    def search_all(self, lc_tokens):
        "Find all terms and starting token index."
        if not self.finalized:
            self.kwtree.finalize()
            self.finalized = True

        zws = self.zws
        content = zws + zws.join(lc_tokens) + zws
        zwsindexes = [i for i, char in enumerate(content) if char == zws]
        results = self.kwtree.search_all(content)

        for result in results:
            # print(f"{result}\n", flush=True)
            t = result[0].strip(zws)
            charpos = result[1]
            index = zwsindexes.index(charpos)
            yield (t, index)
